/*
 * (C) 1997,2003  Jens Ch. Restemeier <jrestemeier@currantbun.com>
 *
 * (C) 2008       Torsten Neuer <tneuer@inwise.de>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *  $Id: 8bf-read.c,v 1.10 2008/06/16 16:34:28 tneuer Exp $
 *
 *  $Log: 8bf-read.c,v $
 *  Revision 1.10  2008/06/16 16:34:28  tneuer
 *  Added guf_filter_(get|set)_edge_mode () access functions
 *
 *  Revision 1.9  2008/06/11 17:51:26  tneuer
 *  Removed assertions not required anymore.
 *  Pointers to const data are const now, too.
 *
 *  Revision 1.8  2008/06/07 16:35:58  tneuer
 *  Using g_strerror () instead of strerror ()
 *
 *  Revision 1.7  2008/06/06 22:43:00  tneuer
 *  Add support for "Gimp::Date" entry in 8bf/afs/ffl/txt file import routines
 *
 *  Revision 1.6  2008/06/05 16:36:04  tneuer
 *  Setting default stepping values for controls (=1) on import
 *
 *  Revision 1.5  2008/06/05 15:23:21  tneuer
 *  Use g_fopen () for UTF-8 compatibility
 *
 *  Revision 1.4  2008/05/23 10:08:43  tneuer
 *  Added command-line filter conversion utilities to compile set
 *
 *  Revision 1.3  2008/05/21 15:24:50  tneuer
 *  Bugfix: code offsets were messaed up.
 *
 *  Revision 1.2  2008/05/21 12:30:23  tneuer
 *  Minor code cleanup.
 *
 *  Revision 1.1.1.1  2008/04/27 13:19:02  tneuer
 *  initial import
 *
 *  Revision 1.1  2008/04/24 14:00:09  tneuer
 *  Initial revision
 *
 */

#include "plugin-conf.h"

#include <libgen.h>
#include <string.h>
#include <errno.h>

#include <glib/gstdio.h>

#include "userfilter.h"
#include "guess-year.h"
#include "8bf.h"
#include "guf.h"


/*
 * This is based on the research and development by 
 * Michael Johannhanwahr <mjohannh@mail.hb.provi.de>.
 * It loads "compiled" Files from FF for Windows.
 */

enum {
  MAGIC_A = 0x06e30,
  MAGIC_B = 0x06c30,
  MAGIC_C = 0x05220,
  MAGIC_D = 0x0cc5a
};

typedef enum FF_VERSION {
  VERSION_A = 0,
  VERSION_B,
  VERSION_C,
  VERSION_D
} FfVersion;

enum {
  VAL_SIZE = 0x0004,       /* size of a value / flag entry */
  LBL_SIZE = 0x0100,       /* max. size of a label entry */
  SRC_SIZE = 0x0400        /* max. size of a source code entry */
};

enum {
  OFS_CATEGORY  = 0x0000,  /* internal offsets in the filter data */
  OFS_TITLE     = 0x0100,  /* structure for reference... */
  OFS_COPYRIGHT = 0x0200,
  OFS_AUTHOR    = 0x0300,
  OFS_MAP_0     = 0x0400,
  OFS_MAP_1     = 0x0500,
  OFS_MAP_2     = 0x0600,
  OFS_MAP_3     = 0x0700,
  OFS_CTL_0     = 0x0800,
  OFS_CTL_1     = 0x0900,
  OFS_CTL_2     = 0x0a00,
  OFS_CTL_3     = 0x0b00,
  OFS_CTL_4     = 0x0c00,
  OFS_CTL_5     = 0x0d00,
  OFS_CTL_6     = 0x0e00,
  OFS_CTL_7     = 0x0f00,
  OFS_SRC_R     = 0x1000,
  OFS_SRC_G     = 0x1400,
  OFS_SRC_B     = 0x1800,
  OFS_SRC_A     = 0x1c00
};

typedef struct FILTER_DATA
{
  gchar category[LBL_SIZE];
  gchar title[LBL_SIZE];
  gchar copyright[LBL_SIZE];
  gchar author[LBL_SIZE];
  gchar map_lbl[N_MAPS][LBL_SIZE];
  gchar ctl_lbl[N_CONTROLS][LBL_SIZE];
  gchar src[N_CHANNELS][SRC_SIZE];
} FilterData;

typedef struct CONTROL_DATA
{
  guchar val[N_CONTROLS][VAL_SIZE];
} ControlData;


static gchar *
clean_string (gchar * s)
{
  gchar * p;

  for (p=s; (*p); p++)
    {
      if (g_ascii_iscntrl(*p))
	{
	  *p = ' ';
	}
    }
  return g_strstrip (s);
}


static gboolean
load_bin_file (FILE *              f,
	       const gchar * const filename,
	       UserFilter *        filter,
	       const FfVersion     ver)
{
  FilterData *  buf = NULL;
  ControlData * val = NULL;
  ControlData * ctl = NULL;
  gboolean      ok  = FALSE;
  gint          i;

  /*
   *  Read control enables...
   */
  switch (ver)
    {
    case VERSION_A:
      fseek (f, 0x7a5c, SEEK_SET);
      break;
    case VERSION_B:
      fseek (f, 0x7a10, SEEK_SET);
      break;
    case VERSION_C:
      fseek (f, 0x5e5c, SEEK_SET);
      break;
    case VERSION_D:
      fseek (f, 0x781C, SEEK_SET);
      break;
    }

  /*
   *  FIXME:  where are the map enables in the .8bf file ?
   */
  ctl = g_new (ControlData, 1);

  if (!fread (ctl, sizeof (ControlData), 1, f))
    {
      goto bailout;
    }
  /*
   *  Read control presets...
   */
  switch (ver)
    {
    case VERSION_A:
      fseek (f, 0x7a1c, SEEK_SET);
      break;
    case VERSION_B:
      fseek (f, 0x79ec, SEEK_SET);
      break;
    case VERSION_C:
      fseek (f, 0x5e1c, SEEK_SET);
      break;
    case VERSION_D:
      fseek (f, 0x77DC, SEEK_SET);
      break;
    }

  val = g_new (ControlData, 1);

  if (!fread (val, sizeof (ControlData), 1, f))
    {
      goto bailout;
    }
  /*
   *  Read filter data...
   */
  switch (ver)
    {
    case VERSION_A:
      fseek (f, 0x7a7c, SEEK_SET);
      break;
    case VERSION_B:
      fseek (f, 0x7a4c, SEEK_SET);
      break;
    case VERSION_C:
      fseek (f, 0x5e7c, SEEK_SET);
      break;
    case VERSION_D:
      fseek (f, 0x783c, SEEK_SET);
      break;
    }

  buf = g_new (FilterData, 1);

  if (fread (buf, sizeof (FilterData), 1, f))
    {
      /*
       *  GUF Header
       *
       *  8BF imports are all level-1 protocol files by default since
       *  they won't have anything fancy.
       */
      guf_filter_set_protocol (filter, GUF_LEVEL_1);
      
      /*
       *  Common information...
       */
      guf_filter_set_imported_category (filter, clean_string (buf->category));
      guf_filter_set_imported_title (filter, clean_string (buf->title));
      guf_filter_set_imported_copyright (filter, clean_string (buf->copyright));
      guf_filter_set_imported_author (filter, clean_string (buf->author));
      
      /*
       *  Filter version information (User Filter extension)...
       */
      guf_filter_set_version_all (filter,1,0,0);
      
      /*
       *  Filter Factory only information (for backward compatibility)...
       */
      g_free (guf_filter_set_imported_8bf_name (filter, g_path_get_basename (filename)));

      /*
       *  Gimp only information...
       */
      guf_filter_set_registered (filter, FALSE);
      guf_filter_set_description (filter, "Imported 8BF filter.");
      guf_filter_set_edge_mode (filter, GIMP_PIXEL_FETCHER_EDGE_SMEAR);
      guf_filter_set_date (filter, guess_year (buf->copyright, filename), 1, 1);

      /*
       *  Controllers...
       */
      for (i = 0; i < N_CONTROLS; i++)
	{
	  if (guf_filter_set_ctl_enabled (filter, i, (gboolean)(ctl->val[i][0])))
	    {
	      guf_filter_set_imported_ctl_label (filter, i, clean_string (buf->ctl_lbl[i]));
	    }
	  guf_filter_set_ctl_preset (filter, i, (gint)(val->val[i][0]));
	  guf_filter_set_ctl_step (filter, i, 1);
	}
      
      /*
       *  Maps...
       *
       *  FIXME:  Where are the map enable entries in the .8bf file ?
       *          We need to do some fixup here unless we have read them
       *          correctly...
       */
      for (i = 0; i < N_MAPS; i++)
	{
	  clean_string (buf->map_lbl[i]);
	  if (guf_filter_set_map_enabled (filter, i,
					  (strlen (buf->map_lbl[i]) > 0) &&
					  guf_filter_get_ctl_enabled (filter, 2 * i) &&
					  guf_filter_get_ctl_enabled (filter, 2 * i + 1)))
	    {
	      guf_filter_set_imported_map_label (filter, i, buf->map_lbl[i]);
	    }
	}

      /*
       *  Source code...
       */
      for (i = 0; i < N_CHANNELS; i++)
	{
	  guf_filter_set_imported_src (filter, i, clean_string (buf->src[i]));
	}
      
      ok = TRUE;
    }

 bailout:
  g_free (buf);
  g_free (val);
  g_free (ctl);
  fclose (f);
  if (!ok)
    {
      g_warning ("8BF %s '%s': %s\n",
		 _("Error loading"), filename, _("File read error."));
    }
  return ok;
}


static gboolean
check_8bf (const gchar * const filename,
	   FILE *              f,
	   const long          fpos)
{
  gchar buf[14];

  if (fseek (f, fpos, SEEK_SET))
    {
      g_warning ("8BF %s '%s': %s\n",
		 _("Error loading"),  filename, g_strerror(errno));
      return FALSE;
    }
  if (fread (buf, sizeof(buf), 1, f) == 1)
    {
      return (strncmp (buf, "Filter Factory", sizeof (buf)) == 0);
    }
  g_warning ("8BF %s '%s': %s\n",
	     _("Error loading"), filename, _("File read error."));
  return FALSE;
}


gboolean
import_8bf (const gchar * const filename,
	    UserFilter *        filter)
{
  FILE * f;
  gchar  buf[2];

  if (!(f = g_fopen (filename, "rb")))
    {
      g_warning ("8BF %s '%s': %s\n",
		 _("Error loading"), filename, g_strerror(errno));
      return FALSE;
    }
  if ( (fread (buf, sizeof (buf), 1, f) == 1) &&
       (strncmp (buf, "MZ", sizeof (buf)) == 0) )
    {
      if (check_8bf (filename, f, MAGIC_A))
	{
	  gchar buf[8];

	  fseek (f, MAGIC_D, SEEK_SET);
	  if (fread (buf, sizeof (buf), 1, f))
	    {
	      const FfVersion v = ( !strncmp (buf, "ESPRESSO", sizeof (buf)) ?
				    VERSION_A : 
				    VERSION_B);

	      return load_bin_file (f, filename, filter, v);
	    }
	}
      else if (check_8bf (filename, f, MAGIC_B))
	{
	  return load_bin_file (f, filename, filter, VERSION_D);
	}
      else if (check_8bf (filename, f, MAGIC_C))
	{
	  return load_bin_file (f, filename, filter, VERSION_C);
	}
      g_warning ("8BF %s '%s': %s\n",
		 _("Error loading"), filename, _("Windows binary is not a FF filter."));
      fclose (f);
      return FALSE;
    }
  g_warning ("8BF %s '%s': %s\n",
	     _("Error loading"), filename, _("Invalid type of file."));
  fclose (f);
  return FALSE;
}

