/*
 * (C) 1997,2003  Jens Ch. Restemeier <jrestemeier@currantbun.com>
 *
 * (C) 2008       Torsten Neuer <tneuer@inwise.de>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *  $Id: afs-read.c,v 1.8 2008/06/16 16:34:32 tneuer Exp $
 *
 *  $Log: afs-read.c,v $
 *  Revision 1.8  2008/06/16 16:34:32  tneuer
 *  Added guf_filter_(get|set)_edge_mode () access functions
 *
 *  Revision 1.7  2008/06/11 17:51:34  tneuer
 *  Removed assertions not required anymore.
 *  Pointers to const data are const now, too.
 *
 *  Revision 1.6  2008/06/07 16:35:58  tneuer
 *  Using g_strerror () instead of strerror ()
 *
 *  Revision 1.5  2008/06/06 22:43:08  tneuer
 *  Add support for "Gimp::Date" entry in 8bf/afs/ffl/txt file import routines
 *
 *  Revision 1.4  2008/06/05 16:36:12  tneuer
 *  Setting default stepping values for controls (=1) on import
 *
 *  Revision 1.3  2008/06/05 15:23:32  tneuer
 *  Use g_fopen () for UTF-8 compatibility
 *
 *  Revision 1.2  2008/05/21 12:30:25  tneuer
 *  Minor code cleanup.
 *
 *  Revision 1.1.1.1  2008/04/27 13:19:05  tneuer
 *  initial import
 *
 *  Revision 1.1  2008/04/24 15:21:33  tneuer
 *  Initial revision
 *
 */

#include "plugin-conf.h"

#include <errno.h>
#include <stdio.h>
#include <libgen.h>
#include <string.h>
#include <stdlib.h>

#include <glib/gstdio.h>

#include "afs.h"
#include "guf.h"
#include "userfilter.h"
#include "guess-year.h"


typedef struct AFS_DATA
{
  gint  val[N_CONTROLS];
  gchar src[N_CHANNELS][8192];
} AfsData;


/*
 * This loads original FF-sourcecode.
 */

static gboolean
load_source_afs (FILE *              f, 
		 const gchar * const filename,
		 AfsData *           filter)
{
  gchar buffer[8192];
  gint  i;
  gint  c;
  gint  n;

  /*
   *  We already read a bit, so we need to rewind...
   */
  fseek (f, 0, SEEK_SET);

  /*
   *  Skip header
   */
  for (c = 0; (((i = fgetc (f)) != '\r') && (c < sizeof (buffer))); )
    {
      buffer[c++] = i;
      if (feof (f))
	{
	  return FALSE;
	}
    }
  if (c == sizeof (buffer))
    {
      return FALSE;
    }
  buffer[c] = '\0';

  /*
   *  Read controls
   */
  for (n = 0; n < N_CONTROLS; n++)
    {
      for (c = 0; (((i = fgetc (f)) != '\r') && (c < sizeof (buffer)));)
	{
	  buffer[c++] = i;
	  if (feof (f))
	    {
	      return FALSE;
	    }
	}
      if (c == sizeof (buffer))
	{
	  return FALSE;
	}
      buffer[c] = '\0';
      filter->val[n] = atoi (buffer);
    }

  /*
   *  Read channels
   */
  for (n = 0; n < N_CHANNELS; n++)
    {
      c = 0;
      do
	{
	  i = fgetc (f);
	  /*
	   * one \r -> skip, two \r -> end of line
	   */
	  if (i == '\r')
	    {
	      i = fgetc (f);
	    }
	  if (i != '\r')
	    {
	      buffer[c++] = i;
	    }
	  if (feof (f))
	    {
	      return FALSE;
	    }
	}
      while ((i != '\r') && (c < sizeof (buffer)));
      if (c == sizeof (buffer))
	{
	  return FALSE;
	}
      buffer[c] = '\0';
      strcpy (filter->src[n], buffer);
    }
  return TRUE;
}


gboolean
import_afs (const gchar * const filename,
	    UserFilter *        filter)
{
  FILE * f;
  gchar  buf[4];

  if (!(f = g_fopen (filename, "r")))
    {
      g_warning ("AFS %s '%s': %s\n",
		 _("Error loading"), filename, g_strerror(errno));
      return FALSE;
    }
  
  if ( (fread (buf, sizeof (buf), 1, f) == 1) &&
       (strncmp (buf, "%RGB", sizeof (buf)) == 0) )
    {
      AfsData * afs;

      afs = g_new (AfsData, 1);
      /*
       * Assume type is .afs
       *
       *  .afs do not have category and title set, so
       *  we set a generic category and derive the title
       *  from the filename...
       */
      if (load_source_afs (f, filename, afs) )
	{
	  gint i;

	  /*
	   *  GUF Header
	   *
	   *  AFS imports are all level-1 protocol files by default since
	   *  they won't have anything fancy.
	   */
	  guf_filter_set_protocol (filter, GUF_LEVEL_1);
	  
	  /*
	   *  Common information...
	   */
	  guf_filter_set_imported_category (filter, "Filter Factory");
	  g_free (guf_filter_set_title (filter, g_path_get_basename(filename)));
	  g_free (guf_filter_set_copyright (filter, 
					    g_strdup_printf("Copyright (C) %s",
							    _("Unknown"))));
	  guf_filter_set_author(filter, _("Unknown"));
	  
	  /*
	   *  Filter version information (User Filter extension)...
	   */
	  guf_filter_set_version_all (filter, 1, 0, 0);

	  /*
	   *  Filter Factory only information (for backward compatibility)...
	   */
	    
	  g_free( guf_filter_set_8bf_name(filter, g_path_get_basename(filename)));
	  /*
	   *  Gimp only information...
	   */
	  guf_filter_set_registered (filter, FALSE);
	  guf_filter_set_description (filter, "Imported AFS filter.");
	  guf_filter_set_date (filter, guess_year ("", filename), 1, 1);
	  guf_filter_set_edge_mode (filter, GIMP_PIXEL_FETCHER_EDGE_SMEAR);

	  /*
	   *  Maps (.afs does not know there either)...
	   */
	  for (i = 0; i < N_MAPS; i++)
	    {
	      guf_filter_set_map_enabled (filter, i, FALSE);
	    }

	  /*
	   *  Controllers...
	   */
	  for (i = 0; i < N_CONTROLS; i++)
	    {
	      guf_filter_set_ctl_enabled (filter, i, TRUE);
	      g_free (guf_filter_set_ctl_label (filter, i,
						g_strdup_printf ("Control %d", i)));
	      guf_filter_set_ctl_preset (filter, i, afs->val[i]);
	      guf_filter_set_ctl_step (filter, i, 1);
	    }
      
	  /*
	   *  Source code...
	   */
	  for (i = 0; i < N_CHANNELS; i++)
	    {
	      guf_filter_set_imported_src (filter, i, g_strstrip(afs->src[i]));
	    }

	  g_free (afs);
	  fclose (f);
	  return TRUE;
	}
      g_free (afs);
    }
  g_warning ("AFS %s '%s': %s\n",
	     _("Error loading"), filename, _("Invalid type of file."));
  fclose (f);
  return FALSE;
}
