/*
 * (C) 1997,2003  Jens Ch. Restemeier <jrestemeier@currantbun.com>
 *
 * (C) 2008       Torsten Neuer <tneuer@inwise.de>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *  $Id: ffl-read.c,v 1.7 2008/06/16 16:34:34 tneuer Exp $
 *
 *  $Log: ffl-read.c,v $
 *  Revision 1.7  2008/06/16 16:34:34  tneuer
 *  Added guf_filter_(get|set)_edge_mode () access functions
 *
 *  Revision 1.6  2008/06/11 17:51:43  tneuer
 *  Removed assertions not required anymore.
 *  Pointers to const data are const now, too.
 *
 *  Revision 1.5  2008/06/06 22:43:09  tneuer
 *  Add support for "Gimp::Date" entry in 8bf/afs/ffl/txt file import routines
 *
 *  Revision 1.4  2008/06/05 16:36:15  tneuer
 *  Setting default stepping values for controls (=1) on import
 *
 *  Revision 1.3  2008/06/05 15:23:49  tneuer
 *  Use g_fopen () for UTF-8 compatibility
 *
 *  Revision 1.2  2008/05/21 12:30:29  tneuer
 *  Minor code cleanup.
 *
 *  Revision 1.1.1.1  2008/04/27 13:19:05  tneuer
 *  initial import
 *
 *  Revision 1.1  2008/04/22 15:38:53  tneuer
 *  Initial revision
 *
 */

#include "plugin-conf.h"

#include <string.h>
#include <stdlib.h>

#include <glib/gstdio.h>

#include "userfilter.h"
#include "guess-year.h"
#include "ffl.h"
#include "guf.h"


/*
 *  Read function for library operations:
 *
 *  Read a string (s) of length (l) from file (f).
 *
 *  Return the actual length of (s) or -1 on unexpected
 *  end-of-file.
 */

static gint
read_string (gchar *      s,
	     const size_t l,
	     FILE *       f)
{
  gint t;

  if (!l)
    {
      *s = '\0';
      return 0;
    }

  /*
   *  We expect that there is more to read, so if it is not,
   *  we get back with an error...
   */
  if (feof (f))
    return -1;

  t = 0;
  do
    {
      gint c;

      if ((c = fgetc (f)) == EOF)
	{
	  s[t] = '\0';
	  return t;
	}

      /*
       *  DOS end-of-line
       *  (this is what we expect with those files as
       *  they come from the DOS world...)
       */
      if (c == '\r')
	{
	  c = fgetc (f);
	  if (c != '\n')
	    ungetc (c, f);
	  s[t] = '\0';
	  return t;
	}

      /*
       *  Reverse DOS end-of-line
       *  (also covers Unix text files - it is here in case
       *  someone messed with the files...)
       */
      if (c == '\n')
	{
	  c = fgetc (f);
	  if (c != '\r')
	    ungetc (c, f);
	  s[t] = '\0';
	  return t;
	}
      s[t++] = c;
    }
  while ((t + 1) < l);

  s[t] = '\0';

  return t;
}


/*
 *  Open library file (s).
 *  Return library file information or NULL.
 */

FFL_info *
ffl_open (const gchar * const filename)
{
  FFL_info * f;
  gchar      buf[10];

  if ((f = g_new (FFL_info, 1)) != NULL)
    {
      if (!(f->f = g_fopen (filename, "rb")))
	{
	  g_free (f);
	  return NULL;		/* file open error */
	}

      if (((read_string (buf, sizeof (buf), f->f) < 6) ||	/* check if FFL */
	   strcmp (buf, "FFL1.0")) ||
	  ((read_string (buf, sizeof (buf), f->f) < 1) ||	/* get filter count */
	   !(f->nfilters = atoi (buf))))
	{
	  fclose (f->f);
	  g_free (f);
	  return NULL;		/* not FFL or empty library */
	}
    }

  return f;
}


/*
 *  Read one library module (data) from library (f)
 *  at file position (fpos).
 *
 *  Return TRUE on success, FALSE on failure.
 */

#define BUFLEN 8192


gboolean
ffl_read (FFL_info *   ffl,
	  long         fpos,
	  UserFilter * filter,
	  gchar *      filename)
{
  gchar * buf;
  gint    i;

  buf = g_new (gchar, BUFLEN);
  fseek (ffl->f, fpos, SEEK_SET);

  /*
   *  GUF Header
   *
   *  FFL imports are all level-1 protocol files by default since
   *  they won't have anything fancy.
   */
  guf_filter_set_protocol (filter, GUF_LEVEL_1);

  /* filename entry */
  if (read_string (buf, BUFLEN, ffl->f) < 1)
    {
      goto abort;
    }

  /*
   *  This takes the "filename" of the library module
   *  as the current filename.
   */
  if (filename)
    {
      strcpy (filename, g_strstrip (buf));
    }

  /*
   *  Common information...
   */
  if (read_string (buf, BUFLEN, ffl->f) < 1)
    {
      goto abort;
    }
  guf_filter_set_imported_category (filter, g_strstrip (buf));
  
  if (read_string (buf, BUFLEN, ffl->f) < 1)
    {
      goto abort;
    }
  guf_filter_set_imported_title (filter, g_strstrip (buf));

  if (read_string (buf, BUFLEN, ffl->f) < 1)
    {
      goto abort;
    }
  guf_filter_set_imported_author (filter, g_strstrip (buf));

  if (read_string (buf, BUFLEN, ffl->f) < 1)
    {
      goto abort;
    }
  guf_filter_set_imported_copyright (filter, g_strstrip (buf));

  /*
   *  Filter version information (User Filter extension)...
   */
  guf_filter_set_version_all (filter, 1, 0, 0);

  /*
   *  Filter Factory only information (for backward compatibility)...
   */
  guf_filter_set_8bf_name (filter, filename);

  /*
   *  Gimp only information...
   */
  guf_filter_set_registered (filter, FALSE);
  guf_filter_set_description (filter, "Imported FFL filter.");
  guf_filter_set_date (filter, guess_year (buf, filename), 1, 1);
  guf_filter_set_edge_mode (filter, GIMP_PIXEL_FETCHER_EDGE_SMEAR);

  /*
   *  Maps...
   */
  for (i = 0; i < N_MAPS; i++)
    {
      if (read_string (buf, BUFLEN, ffl->f) < 0)
	{
	  goto abort;
	}
      g_strstrip (buf);
      if (guf_filter_set_map_enabled (filter, i, (strlen (buf) > 0)))
	{
	  guf_filter_set_imported_map_label (filter, i, buf);
	}
    }

  /*
   *  Controllers...
   */
  for (i = 0; i < N_CONTROLS; i++)
    {
      if (read_string (buf, BUFLEN, ffl->f) < 0)
	{
	  goto abort;
	}
      g_strstrip (buf);
      if (guf_filter_set_ctl_enabled (filter, i, (strlen (buf) > 0)))
	{
	  guf_filter_set_imported_ctl_label (filter, i, buf);
	}
    }

  for (i = 0; i < N_CONTROLS; i++)
    {
      gint    n;
      
      if (read_string (buf, BUFLEN, ffl->f) < 1)
	{
	  goto abort;
	}
      n = atoi (buf);
      guf_filter_set_ctl_preset (filter, i, (n < 0 ? 0 : n > 255 ? 255 : n));
      guf_filter_set_ctl_step (filter, i, 1);
    }

  /*
   *  Source code...
   */
  for (i = 0; i < N_CHANNELS; i++)
    {
      if (read_string (buf, BUFLEN, ffl->f) < 1)
	{
	  goto abort;
	}
      guf_filter_set_imported_src (filter, i, g_strstrip(buf));
    }

  g_free (buf);
  return TRUE;

abort:
  g_free (buf);
  return FALSE;
}


/*
 *  Close library file (f).
 */

gint
ffl_close (FFL_info * ffl)
{
  gint v;

  v = fclose (ffl->f);
  g_free (ffl);
  return v;
}
