/*
 * (C) 2008  Torsten Neuer <tneuer@inwise.de>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *  $Id: guf-filter-get.c,v 1.11 2008/06/16 16:34:35 tneuer Exp $
 *
 *  $Log: guf-filter-get.c,v $
 *  Revision 1.11  2008/06/16 16:34:35  tneuer
 *  Added guf_filter_(get|set)_edge_mode () access functions
 *
 *  Revision 1.10  2008/06/11 17:51:53  tneuer
 *  Removed assertions not required anymore.
 *  Pointers to const data are const now, too.
 *
 *  Revision 1.9  2008/06/07 20:06:06  tneuer
 *  Use __PRETTY_FUNCTION__ for function names in warnings
 *
 *  Revision 1.8  2008/06/05 14:39:14  tneuer
 *  Added query functions for "Gimp::Date" and "Control %u::Step"
 *
 *  Revision 1.7  2008/06/05 09:41:51  tneuer
 *  Using glib functions for directory scanning instead of <dirent.h> and <stat.h>
 *
 *  Revision 1.6  2008/06/03 15:17:33  tneuer
 *  Bugfix: wrong locale of labels / tooltips in the filter control panel when re-running the plug-in
 *  Bugfix: wrong state of the filter editor panel when re-running the plug-in
 *
 *  Revision 1.5  2008/06/03 11:32:49  tneuer
 *  Bugfix: filter control panel does not display locale for tooltips
 *
 *  Revision 1.4  2008/05/26 15:10:54  tneuer
 *  Minor code cleanup.
 *
 *  Revision 1.3  2008/05/21 15:26:26  tneuer
 *  Use default fallback value for localized items.
 *
 *  Revision 1.2  2008/05/05 13:32:40  tneuer
 *  New functions guf_filter_renew_from_pdb () and guf_filter_dump () for PDB operations.
 *
 *  Revision 1.1.1.1  2008/04/27 13:19:04  tneuer
 *  initial import
 *
 *  Revision 1.1  2008/04/25 13:46:09  tneuer
 *  Initial revision
 *
 */

#include "plugin-conf.h"

#include <stdlib.h>
#include <time.h>

#include <glib.h>

#include "userfilter.h"
#include "guf.h"
#include "8bf.h"


static const gchar * const error_format = "%s (): Fallback to default value: %s";

/*
 *  Primitive functions to get filter data from the internal
 *  representation.
 *
 *  Function names are self-explanatory.
 *  Functions returning (gchar *) will allocate a new memory
 *  region which must be g_free ()'d again.
 *
 *  The 'locale' functions will query the preferred locale
 *  and try to return the value in that locale if possible.
 */

gchar *
guf_filter_dump (const UserFilter * filter)
{
  GError * err = NULL;
  gchar  * val;
  gsize    len;

  val = g_key_file_to_data ((GKeyFile *)filter, &len, &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      if (val == NULL)
	{
	  val = g_strdup ("");
	}
      g_error_free (err);
    }
  return val;
}


GufProtocol
guf_filter_get_protocol (const UserFilter * filter)
{
  GError * err = NULL;
  gint     val;

  val = g_key_file_get_integer ((GKeyFile *)filter, "GUF", "Protocol", &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
      return GUF_LEVEL_1;
    }
  return (GufProtocol)val;
}


gboolean
guf_filter_get_registered (const UserFilter * filter)
{
  GError * err = NULL;
  gboolean val;

  val = g_key_file_get_boolean ((GKeyFile *)filter, "Gimp", "Registered", &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
    }
  return val;
}


gchar *
guf_filter_get_title (const UserFilter * filter)
{
  GError * err = NULL;
  gchar *  val;

  val = g_key_file_get_string ((GKeyFile *)filter, "Info", "Title", &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
      val = g_strdup ("Untitled");
    }
  return val;
}


gchar *
guf_filter_get_locale_title (const UserFilter * filter)
{
  GError * err = NULL;
  gchar *  val;

  val = g_key_file_get_locale_string ((GKeyFile *)filter, "Info", "Title", NULL, &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
      if (val == NULL)
	{
	  val = g_strdup (_("Untitled"));
	}
    }
  return val;
}


gchar *
guf_filter_get_author (const UserFilter * filter)
{
  GError * err = NULL;
  gchar *  val;

  val = g_key_file_get_string ((GKeyFile *)filter, "Info", "Author", &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
      val = g_strdup (g_get_real_name ());
    }
  return val;
}


gchar *
guf_filter_get_copyright (const UserFilter * filter)
{
  GError * err = NULL;
  gchar *  val;

  val = g_key_file_get_string ((GKeyFile *)filter, "Info", "Copyright", &err);
  if (err != NULL)
    {
      GDate * date;

      date = g_date_new ();
      g_date_set_time_t (date, time (NULL));
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
      val = g_strdup_printf ("Copyright (C) %u %s",
			     g_date_get_year (date),
			     g_get_real_name ());
      g_date_free (date);
    }
  return val;
}


gchar *
guf_filter_get_category (const UserFilter * filter)
{
  GError * err = NULL;
  gchar *  val;

  val = g_key_file_get_string ((GKeyFile *)filter, "Info", "Category", &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
      val = g_strdup ("<Image>/Filters/User Filters");
    }
  return val;
}


gchar *
guf_filter_get_8bf_name (const UserFilter * filter)
{
  GError * err = NULL;
  gchar *  val;

  val = g_key_file_get_string ((GKeyFile *)filter, "Filter Factory", "8bf", &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
      val = g_strdup_printf ("%s%s", _("untitled"), EXT_8BF);
    }
  return val;
}


gchar *
guf_filter_get_description (const UserFilter * filter)
{
  GError * err = NULL;
  gchar *  val;

  val = g_key_file_get_string ((GKeyFile *)filter, "Gimp", "Description", &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
      val = g_strdup ("Default filter description.");
    }
  return val;
}


gchar *
guf_filter_get_locale_description (const UserFilter * filter)
{
  GError * err = NULL;
  gchar *  val;

  val = g_key_file_get_locale_string ((GKeyFile *)filter, "Gimp", "Description", NULL, &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
      if (val == NULL)
	{
	  val = g_strdup (_("Default filter description."));
	}
    }
  return val;
}


gchar *
guf_filter_get_date_string (const UserFilter * filter)
{
  GError * err = NULL;
  gchar *  val;

  val = g_key_file_get_string ((GKeyFile *)filter, "Gimp", "Date", &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
      if (val == NULL)
	{
	  GDate * date;

	  date = g_date_new ();
	  val = g_strdup_printf ("%4u-%02u-%02u",
				 g_date_get_year (date),
				 g_date_get_month (date),
				 g_date_get_day (date));
	  g_date_free (date);
	}
    }
  return val;
}


void
guf_filter_get_date (const UserFilter * filter,
		     guint *            year,
		     guint *            month,
		     guint *            day)
{
  gchar * date;
  gchar * next = NULL;

  date = guf_filter_get_date_string (filter);

  *year  = strtol (date, &next, 10);
  ++next;
  *month = strtol (next, &next, 10);
  ++next;
  *day   = strtol (next, &next, 10);

  g_free (date);
}


guint
guf_filter_get_date_year (const UserFilter * filter)
{
  guint y, m, d;

  guf_filter_get_date (filter, &y, &m, &d);
  return y;
}


guint
guf_filter_get_date_month (const UserFilter * filter)
{
  guint y, m, d;

  guf_filter_get_date (filter, &y, &m, &d);
  return m;
}


guint
guf_filter_get_date_day (const UserFilter * filter)
{
  guint y, m, d;

  guf_filter_get_date (filter, &y, &m, &d);
  return d;
}


GimpPixelFetcherEdgeMode
guf_filter_get_edge_mode (const UserFilter * filter)
{
  GimpPixelFetcherEdgeMode val;
  GError *                 err = NULL;

  val = g_key_file_get_integer ((GKeyFile *)filter, "Gimp", "EdgeMode", &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
      val = GIMP_PIXEL_FETCHER_EDGE_SMEAR;
    }
  return val;
}


gint
guf_filter_get_version_major (const UserFilter * filter)
{
  GError * err = NULL;
  gint     val;

  val = g_key_file_get_integer ((GKeyFile *)filter, "Version", "Major", &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
    }
  return val;
}


gint
guf_filter_get_version_minor (const UserFilter * filter)
{
  GError * err = NULL;
  gint     val;

  val = g_key_file_get_integer ((GKeyFile *)filter, "Version", "Minor", &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
    }
  return val;
}


gint
guf_filter_get_version_micro (const UserFilter * filter)
{
  GError * err = NULL;
  gint     val;

  val = g_key_file_get_integer ((GKeyFile *)filter, "Version", "Micro", &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
    }
  return val;
}


gchar *
guf_filter_get_version_string (const UserFilter * filter)
{
  return g_strdup_printf ("%d.%d.%d",
			  guf_filter_get_version_major (filter),
			  guf_filter_get_version_minor (filter),
			  guf_filter_get_version_micro (filter));
}


gboolean
guf_filter_get_map_enabled (const UserFilter * filter,
			    const guint        map)
{
  GError * err = NULL;
  gchar *  grp;
  gboolean val;

  grp = g_strdup_printf ("Map %u", map);
  val = g_key_file_get_boolean ((GKeyFile *)filter, grp, "Enabled", &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
    }
  g_free (grp);
  return val;
}


gchar *
guf_filter_get_map_label (const UserFilter * filter,
			  const guint        map)
{
  GError * err = NULL;
  gchar *  grp;
  gchar *  val;

  grp = g_strdup_printf ("Map %u", map);
  val = g_key_file_get_string ((GKeyFile *)filter, grp, "Label", &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
      val = g_strdup (grp);
    }
  g_free (grp);
  return val;
}


gchar *
guf_filter_get_locale_map_label (const UserFilter * filter,
				 const guint        map)
{
  GError * err = NULL;
  gchar *  grp;
  gchar *  val;

  grp = g_strdup_printf ("Map %u", map);
  val = g_key_file_get_locale_string ((GKeyFile *)filter, grp, "Label", NULL, &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
      if (val == NULL)
	{
	  val = g_strdup (grp);
	}
    }
  g_free (grp);
  return val;
}


gchar *
guf_filter_get_map_tooltip (const UserFilter * filter,
			    const guint        map)
{
  GError * err = NULL;
  gchar *  grp;
  gchar *  val;

  grp = g_strdup_printf ("Map %u", map);
  val = g_key_file_get_string ((GKeyFile *)filter, grp, "Tooltip", &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
      val = g_strdup ("Use these combined controls to alter the effects of the filter...");
    }
  g_free (grp);
  return val;
}


gchar *
guf_filter_get_locale_map_tooltip (const UserFilter * filter,
				   const guint        map)
{
  GError * err = NULL;
  gchar *  grp;
  gchar *  val;

  grp = g_strdup_printf ("Map %u", map);
  val = g_key_file_get_locale_string ((GKeyFile *)filter, grp, "Tooltip", NULL, &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
      if (val == NULL)
	{
	  val = g_strdup (_("Use these combined controls to alter the effects of the filter..."));
	}
    }
  g_free (grp);
  return val;
}


gboolean
guf_filter_get_ctl_enabled (const UserFilter * filter,
			    const guint        ctl)
{
  GError * err = NULL;
  gchar *  grp;
  gboolean val;

  grp = g_strdup_printf ("Control %u", ctl);
  val = g_key_file_get_boolean ((GKeyFile *)filter, grp, "Enabled", &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
    }
  g_free (grp);
  return val;
}


gchar *
guf_filter_get_ctl_label (const UserFilter * filter,
			  const guint        ctl)
{
  GError * err = NULL;
  gchar *  grp;
  gchar *  val;

  grp = g_strdup_printf ("Control %u", ctl);
  val = g_key_file_get_string ((GKeyFile *)filter, grp, "Label", &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
      val = g_strdup (grp);
    }
  g_free (grp);
  return val;
}


gchar *
guf_filter_get_locale_ctl_label (const UserFilter * filter,
				 const guint        ctl)
{
  GError * err = NULL;
  gchar *  grp;
  gchar *  val;

  grp = g_strdup_printf ("Control %u", ctl);
  val = g_key_file_get_locale_string ((GKeyFile *)filter, grp, "Label", NULL, &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
      if (val == NULL)
	{
	  val = g_strdup (grp);
	}
    }
  g_free (grp);
  return val;
}


gchar *
guf_filter_get_ctl_tooltip (const UserFilter * filter,
			    const guint        ctl)
{
  GError * err = NULL;
  gchar *  grp;
  gchar *  val;

  grp = g_strdup_printf ("Control %u", ctl);
  val = g_key_file_get_string ((GKeyFile *)filter, grp, "Tooltip", &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
      val = g_strdup ("Use this control to alter the effects of the filter...");
    }
  g_free (grp);
  return val;
}


gchar *
guf_filter_get_locale_ctl_tooltip (const UserFilter * filter,
				   const guint        ctl)
{
  GError * err = NULL;
  gchar *  grp;
  gchar *  val;

  grp = g_strdup_printf ("Control %u", ctl);
  val = g_key_file_get_locale_string ((GKeyFile *)filter, grp, "Tooltip", NULL, &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
      if (val == NULL)
	{
	  val = g_strdup (_("Use this control to alter the effects of the filter..."));
	}
    }
  g_free (grp);
  return val;
}


gint
guf_filter_get_ctl_preset (const UserFilter * filter,
			   const guint        ctl)
{
  GError * err = NULL;
  gchar *  grp;
  gint     val;

  grp = g_strdup_printf ("Control %u", ctl);
  val = g_key_file_get_integer ((GKeyFile *)filter, grp, "Preset", &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
    }
  g_free (grp);
  return val;
}


gint
guf_filter_get_ctl_step (const UserFilter * filter,
			 const guint        ctl)
{
  GError * err = NULL;
  gchar *  grp;
  gint     val;

  grp = g_strdup_printf ("Control %u", ctl);
  val = g_key_file_get_integer ((GKeyFile *)filter, grp, "Step", &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
      val = 1;
    }
  else if ((256 % val) != 0)
    {
      g_warning (error_format, G_STRFUNC, _("Invalid value for stepping"));
      val = 1;
    }
  g_free (grp);
  return val;
}


gchar *
guf_filter_get_src (const UserFilter * filter,
		    const guint        channel)
{
  static const gchar * RGBA[N_CHANNELS] = {
    "R", "G", "B", "A"
  };
  static const gchar * rgba[N_CHANNELS] = {
    "r", "g", "b", "a"
  };
  GError * err = NULL;
  gchar *  val;

  val = g_key_file_get_string ((GKeyFile *)filter, "Code", RGBA[channel], &err);
  if (err != NULL)
    {
      g_warning (error_format, G_STRFUNC, err->message);
      g_error_free (err);
      val = g_strdup (rgba[channel]);
    }
  return val;
}
