/*
 * (C) 2008       Torsten Neuer <tneuer@inwise.de>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *  $Id: guf-filter-renew.c,v 1.13 2008/06/17 13:52:38 tneuer Exp $
 *
 *  $Log: guf-filter-renew.c,v $
 *  Revision 1.13  2008/06/17 13:52:38  tneuer
 *  Bugfix: Add missing default for map tooltips
 *
 *  Revision 1.12  2008/06/16 16:34:35  tneuer
 *  Added guf_filter_(get|set)_edge_mode () access functions
 *
 *  Revision 1.11  2008/06/11 17:51:58  tneuer
 *  Removed assertions not required anymore.
 *  Pointers to const data are const now, too.
 *
 *  Revision 1.10  2008/06/07 20:06:06  tneuer
 *  Use __PRETTY_FUNCTION__ for function names in warnings
 *
 *  Revision 1.9  2008/06/05 13:51:51  tneuer
 *  Set default date and control stepping in default filter.
 *
 *  Revision 1.8  2008/06/05 09:41:59  tneuer
 *  Using glib functions for directory scanning instead of <dirent.h> and <stat.h>
 *
 *  Revision 1.7  2008/06/03 14:19:09  tneuer
 *  Minor code cleanup
 *
 *  Revision 1.6  2008/05/22 11:03:43  tneuer
 *  New callback functions for source and meta-info entry widgets.
 *
 *  Revision 1.5  2008/05/16 16:37:15  tneuer
 *  Minor code cleanup.
 *
 *  Revision 1.4  2008/05/16 14:02:51  tneuer
 *  Loading stand-alone (registered) filters.
 *
 *  Revision 1.3  2008/05/13 15:38:49  tneuer
 *  Userfilter registry run-time fixes.
 *
 *  Revision 1.2  2008/05/05 13:32:40  tneuer
 *  New functions guf_filter_renew_from_pdb () and guf_filter_dump () for PDB operations.
 *
 *  Revision 1.1.1.1  2008/04/27 13:19:05  tneuer
 *  initial import
 *
 *  Revision 1.1  2008/04/26 17:52:45  tneuer
 *  Initial revision
 *
 */

#include "plugin-conf.h"

#include <time.h>

#include <glib.h>

#include "userfilter.h"
#include "registry.h"
#include "guf.h"
#include "8bf.h"


/*
 *  Reset the filter to a default state.
 *
 *  If filter is not NULL, the old filter is destroyed.
 *  In any case, a new filter is created and filled with
 *  default values.
 */

UserFilter *
guf_filter_renew (UserFilter * filter)
{
  GDate * date = g_date_new ();
  guint   i;

  g_date_set_time_t (date, time (NULL));

  if (filter != NULL)
    {
      guf_filter_free (filter);
    }
  filter = guf_filter_new ();

  /*
   *  GUF Header
   */
  guf_filter_set_protocol (filter, GUF_PROTOCOL);
	  
  /*
   *  Common information...
   */
  g_free (guf_filter_set_category (filter,
				   g_strdup_printf ("<Image>/Filters/%s",
						    _("User Filters"))));
  guf_filter_set_title (filter, "Untitled");

  guf_filter_set_author (filter, g_get_real_name ());
  g_free (guf_filter_set_copyright (filter,
				    g_strdup_printf ("Copyright © %u %s",
						     g_date_get_year (date),
						     g_get_real_name ())));

  /*
   *  Filter version information (User Filter extension)...
   */
  guf_filter_set_version_all (filter, 0, 1, 0);

  /*
   *  Filter Factory only information (for backward compatibility)...
   */
  g_free (guf_filter_set_8bf_name (filter, 
				   g_strdup_printf ("%s%s",
						    _("untitled"),
						    EXT_8BF)));

  /*
   *  Gimp only information...
   */
  guf_filter_set_registered (filter, FALSE);
  guf_filter_set_description (filter, "Default filter.");
  guf_filter_set_date (filter,
		       g_date_get_year (date),
		       g_date_get_month (date),
		       g_date_get_day (date));
  guf_filter_set_edge_mode (filter, GIMP_PIXEL_FETCHER_EDGE_SMEAR);

  /*
   *  Maps (.txt does not know them)...
   */
  for (i = 0; i < N_MAPS; i++)
    {
      guf_filter_set_map_enabled (filter, i, FALSE);
      g_free (guf_filter_set_map_label (filter, i,
					g_strdup_printf ("Map %u", i)));
      guf_filter_set_map_tooltip (filter, i,
				  "Use these mapped sliders to change the effects of the filter.");
    }

  /*
   *  Controllers...
   */
  for (i = 0; i < N_CONTROLS; i++)
    {
      guf_filter_set_ctl_enabled (filter, i, TRUE);
      g_free (guf_filter_set_ctl_label (filter, i, 
					g_strdup_printf ("Control %u", i)));
      guf_filter_set_ctl_tooltip (filter, i,
				  "Use this slider to change the effects of the filter.");
      guf_filter_set_ctl_preset (filter, i, 0);
      guf_filter_set_ctl_step (filter, i, 1);
    }
      
  /*
   *  Source code...
   */
  for (i = 0; i < N_CHANNELS; i++ )
    {
      static const gchar * const rgba[N_CHANNELS] = {
	"r", "g", "b", "a"
      };
      guf_filter_set_src (filter, i, rgba[i]);
    }

  g_date_free (date);

  return filter;
}


/*
 *  Load a filter from a Gimp PDB entry.
 *
 *  If filter is not NULL, the old filter is destroyed.
 *  In any case, a new filter is created and filled with
 *  PDB values.
 */

UserFilter *
guf_filter_renew_from_pdb (UserFilter *        filter,
			   const gchar * const data_key)
{
  const gsize len = gimp_get_data_size (data_key);

  if (len > 0)
    {
      GError * err = NULL;
      gchar *  val = g_new (gchar, len);

      if (filter != NULL)
	{
	  guf_filter_free (filter);
	}
      filter = guf_filter_new ();

      gimp_get_data (data_key, val);

      g_key_file_load_from_data ((GKeyFile *)filter, val, len, GUF_FLAGS, &err);
      g_free (val);

      if (err != NULL)
	{
	  /*
	   *  On error, fallback to default filter...
	   */
	  g_warning ("%s (): %s: %s",
		     G_STRFUNC, _("Error loading filter"), err->message);
	  g_error_free (err);
	  filter = guf_filter_renew (filter);
	}
    }

  return filter;
}


/*
 *  Load a filter from the filter file that a pdb-name points to.
 *
 *  If the filter is not NULL, the previous filter will be destroyed.
 *  The filter will be loaded from the file that is registered with
 *  the pdb-name.  If the load fails, it will be initialized to the
 *  default filter values.
 */

UserFilter *
guf_filter_renew_from_registry (UserFilter *        filter,
				const gchar * const rcfile,
				const gchar * const pdbname)
{
  Registry * reg;
  gchar *    path;

  reg = registry_load (registry_new (), rcfile);
  path = registry_get_filter_path (reg, pdbname);
  if (path != NULL)
    {
      if (filter != NULL)
	{
	  guf_filter_free (filter);
	}
      filter = guf_filter_new ();
      if (!guf_load (path, GUF_PROTOCOL, filter))
	{
	  filter = guf_filter_renew (filter);
	}
      g_free (path);
    }
  else
    {
      filter = guf_filter_renew (filter);
    }
  registry_free (reg);

  return filter;
}
