/*
 * (C) 2008  Torsten Neuer <tneuer@inwise.de>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *  $Id: guf-filter-set.c,v 1.8 2008/06/16 16:34:37 tneuer Exp $
 *
 *  $Log: guf-filter-set.c,v $
 *  Revision 1.8  2008/06/16 16:34:37  tneuer
 *  Added guf_filter_(get|set)_edge_mode () access functions
 *
 *  Revision 1.7  2008/06/11 17:52:00  tneuer
 *  Removed assertions not required anymore.
 *  Pointers to const data are const now, too.
 *
 *  Revision 1.6  2008/06/06 09:23:32  tneuer
 *  Build a "correct" Gimp menu path for imported category values
 *
 *  Revision 1.5  2008/06/05 13:46:28  tneuer
 *  Added set-value functions for "Gimp::Date" and "Control %u::Step"
 *
 *  Revision 1.4  2008/05/26 15:10:54  tneuer
 *  Minor code cleanup.
 *
 *  Revision 1.3  2008/05/23 10:08:45  tneuer
 *  Added command-line filter conversion utilities to compile set
 *
 *  Revision 1.2  2008/05/21 12:30:32  tneuer
 *  Minor code cleanup.
 *
 *  Revision 1.1.1.1  2008/04/27 13:19:04  tneuer
 *  initial import
 *
 *  Revision 1.1  2008/04/25 14:32:20  tneuer
 *  Initial revision
 *
 */

#include "plugin-conf.h"

#include <glib.h>

#include "userfilter.h"
#include "guf.h"
#include "8bf.h"


static const gchar *
guf_import_string (UserFilter *  filter,
		   const gchar * const grp,
		   const gchar * const key,
		   const gchar * const val)
{
  GError * err = NULL;
  gsize    r;
  gsize    w;
  gchar *  t = NULL;

  if ( (t = g_convert (val, -1, "UTF-8", "WINDOWS-1252", &r, &w, &err)) == NULL )
    {
      g_warning ( "%s: %s\n", _("Unable to convert data"), err->message );
      g_error_free (err);
      g_free (t);
      t = g_strdup (val);
    }
  g_key_file_set_string ((GKeyFile *)filter, grp, key, t);
  g_free (t);
  return val;
}


/*
 *  Primitive functions to set filter data in the internal
 *  representation.
 *
 *  Function names are self-explanatory.
 *
 *  The "imported" functions will attempt conversion to utf-8
 *  characterset (used by non-native file readers).
 *
 *  The "locale" functions will set an entry for the specified
 *  locale value.
 */

void
guf_filter_set_protocol (UserFilter *      filter,
			 const GufProtocol value)
{
  g_key_file_set_integer ((GKeyFile *)filter, "GUF", "Protocol", (gint)value);
}


void
guf_filter_set_registered (UserFilter *   filter,
			   const gboolean value)
{
  g_key_file_set_boolean ((GKeyFile *)filter, "Gimp", "Registered", value);
}


gchar *
guf_filter_set_title (UserFilter *        filter,
		      const gchar * const value)
{
  g_key_file_set_string ((GKeyFile *)filter, "Info", "Title", value);
  return (gchar *)value;
}


gchar *
guf_filter_set_locale_title (UserFilter *  filter,
			     const gchar * const locale,
			     const gchar * const value)
{
  g_key_file_set_locale_string ((GKeyFile *)filter, "Info", "Title", locale, value);
  return (gchar *)value;
}


gchar *
guf_filter_set_imported_title (UserFilter *        filter,
			       const gchar * const value)
{
  return (gchar *)guf_import_string (filter, "Info", "Title", value);
}


gchar *
guf_filter_set_author (UserFilter *        filter,
		       const gchar * const value)
{
  g_key_file_set_string ((GKeyFile *)filter, "Info", "Author", value);
  return (gchar *)value;
}


gchar *
guf_filter_set_imported_author (UserFilter *        filter,
				const gchar * const value)
{
  return (gchar *)guf_import_string (filter, "Info", "Author", value);
}


gchar *
guf_filter_set_copyright (UserFilter *        filter,
			  const gchar * const value)
{
  g_key_file_set_string ((GKeyFile *)filter, "Info", "Copyright", value);
  return (gchar *)value;
}


gchar *
guf_filter_set_imported_copyright (UserFilter *        filter,
				   const gchar * const value)
{
  return (gchar *)guf_import_string (filter, "Info", "Copyright", value);
}


gchar *
guf_filter_set_category (UserFilter *        filter,
			 const gchar * const value)
{
  g_key_file_set_string ((GKeyFile *)filter, "Info", "Category", value);
  return (gchar *)value;
}


gchar *
guf_filter_set_imported_category (UserFilter *        filter,
				  const gchar * const value)
{
  gchar * category;

  category = g_strconcat ("<Image>/Filters/", value, NULL);
  guf_import_string (filter, "Info", "Category", category);
  g_free (category);
  return (gchar *)value;
}


gchar *
guf_filter_set_8bf_name (UserFilter *        filter,
			 const gchar * const value)
{
  g_key_file_set_string ((GKeyFile *)filter, "Filter Factory", "8bf", value);
  return (gchar *)value;
}


gchar *
guf_filter_set_imported_8bf_name (UserFilter *        filter,
				  const gchar * const value)
{
  return (gchar *)guf_import_string (filter, "Filter Factory", "8bf", value);
}


gchar *
guf_filter_set_description (UserFilter *        filter,
			    const gchar * const value)
{
  g_key_file_set_string ((GKeyFile *)filter, "Gimp", "Description", value);
  return (gchar *)value;
}


gchar *
guf_filter_set_locale_description (UserFilter *        filter,
				   const gchar * const locale,
				   const gchar * const value)
{
  g_key_file_set_locale_string ((GKeyFile *)filter, "Gimp", "Description", locale, value);
  return (gchar *)value;
}


gchar *
guf_filter_set_date_string (UserFilter *        filter,
			    const gchar * const date)
{
  g_key_file_set_string ((GKeyFile *)filter, "Gimp", "Date", date);
  return (gchar *)date;
}


void
guf_filter_set_date (UserFilter * filter,
		     const guint  year,
		     const guint  month,
		     const guint  day)
{
  g_free (guf_filter_set_date_string (filter,
				      g_strdup_printf ("%4u-%02u-%02u",
						       year, month, day)));
}


GimpPixelFetcherEdgeMode
guf_filter_set_edge_mode (UserFilter *             filter,
			  GimpPixelFetcherEdgeMode value)
{
  g_key_file_set_integer ((GKeyFile *)filter, "Gimp", "EdgeMode", value);
  return value;
}


void
guf_filter_set_version_major (UserFilter * filter,
			      const gint   value)
{
  g_key_file_set_integer ((GKeyFile *)filter, "Version", "Major", value);
}


void
guf_filter_set_version_minor (UserFilter * filter,
			      const gint   value)
{
  g_key_file_set_integer ((GKeyFile *)filter, "Version", "Minor", value);
}


void
guf_filter_set_version_micro (UserFilter * filter,
			      const gint   value)
{
  g_key_file_set_integer ((GKeyFile *)filter, "Version", "Micro", value);
}


void
guf_filter_set_version_all (UserFilter * filter,
			    const gint   major,
			    const gint   minor,
			    const gint   micro)
{
  guf_filter_set_version_major (filter, major);
  guf_filter_set_version_minor (filter, minor);
  guf_filter_set_version_micro (filter, micro);
}


gboolean
guf_filter_set_map_enabled (UserFilter *   filter,
			    const guint    map,
			    const gboolean value)
{
  gchar * grp;

  grp = g_strdup_printf ("Map %u", map);
  g_key_file_set_boolean ((GKeyFile *)filter, grp, "Enabled", value);
  g_free (grp);
  return value;
}


gchar *
guf_filter_set_map_label (UserFilter *        filter,
			  const guint         map,
			  const gchar * const value)
{
  gchar * grp;

  grp = g_strdup_printf ("Map %u", map);
  g_key_file_set_string ((GKeyFile *)filter, grp, "Label", value);
  g_free (grp);
  return (gchar *)value;
}


gchar *
guf_filter_set_locale_map_label (UserFilter *        filter,
				 const guint         map,
				 const gchar * const locale,
				 const gchar * const value)
{
  gchar * grp;

  grp = g_strdup_printf ("Map %u", map);
  g_key_file_set_locale_string ((GKeyFile *)filter, grp, "Label", locale, value);
  g_free (grp);
  return (gchar *)value;
}


gchar *
guf_filter_set_imported_map_label (UserFilter *        filter,
				   const guint         map,
				   const gchar * const value)
{
  gchar * grp;

  grp = g_strdup_printf ("Map %u", map);
  guf_import_string (filter, grp, "Label", value);
  g_free (grp);
  return (gchar *)value;
}


gchar *
guf_filter_set_map_tooltip (UserFilter *        filter,
			    const guint         map,
			    const gchar * const value)
{
  gchar * grp;

  grp = g_strdup_printf ("Map %u", map);
  g_key_file_set_string ((GKeyFile *)filter, grp, "Tooltip", value);
  g_free (grp);
  return (gchar *)value;
}


gchar *
guf_filter_set_locale_map_tooltip (UserFilter *        filter,
				   const guint         map,
				   const gchar * const locale,
				   const gchar * const value)
{
  gchar * grp;

  grp = g_strdup_printf ("Map %u", map);
  g_key_file_set_locale_string ((GKeyFile *)filter, grp, "Tooltip", locale, value);
  g_free (grp);
  return (gchar *)value;
}


gboolean
guf_filter_set_ctl_enabled (UserFilter *   filter,
			    const guint    ctl,
			    const gboolean value)
{
  gchar * grp;

  grp = g_strdup_printf ("Control %u", ctl);
  g_key_file_set_boolean ((GKeyFile *)filter, grp, "Enabled", value);
  g_free (grp);
  return value;
}


gchar *
guf_filter_set_ctl_label (UserFilter *        filter,
			  const guint         ctl,
			  const gchar * const value)
{
  gchar * grp;

  grp = g_strdup_printf ("Control %u", ctl);
  g_key_file_set_string ((GKeyFile *)filter, grp, "Label", value);
  g_free (grp);
  return (gchar *)value;
}


gchar *
guf_filter_set_locale_ctl_label (UserFilter *        filter,
				 const guint         ctl,
				 const gchar * const locale,
				 const gchar * const value)
{
  gchar * grp;

  grp = g_strdup_printf ("Control %u", ctl);
  g_key_file_set_locale_string ((GKeyFile *)filter, grp, "Label", locale, value);
  g_free (grp);
  return (gchar *)value;
}


gchar *
guf_filter_set_imported_ctl_label (UserFilter *        filter,
				   const guint         ctl,
				   const gchar * const value)
{
  gchar * grp;

  grp = g_strdup_printf ("Control %u", ctl);
  guf_import_string (filter, grp, "Label", value);
  g_free (grp);
  return (gchar *)value;
}


gchar *
guf_filter_set_ctl_tooltip (UserFilter *        filter,
			    const guint         ctl,
			    const gchar * const value)
{
  gchar * grp;

  grp = g_strdup_printf ("Control %u", ctl);
  g_key_file_set_string ((GKeyFile *)filter, grp, "Tooltip", value);
  g_free (grp);
  return (gchar *)value;
}


gchar *
guf_filter_set_locale_ctl_tooltip (UserFilter *        filter,
				   const guint         ctl,
				   const gchar * const locale,
				   const gchar * const value)
{
  gchar * grp;

  grp = g_strdup_printf ("Control %u", ctl);
  g_key_file_set_locale_string ((GKeyFile *)filter, grp, "Tooltip", locale, value);
  g_free (grp);
  return (gchar *)value;
}


void
guf_filter_set_ctl_preset (UserFilter * filter,
			   const guint  ctl,
			   const gint   value)
{
  gchar * grp;

  grp = g_strdup_printf ("Control %u", ctl);
  g_key_file_set_integer ((GKeyFile *)filter, grp, "Preset", value);
  g_free (grp);
}


void
guf_filter_set_ctl_step (UserFilter * filter,
			 const guint  ctl,
			 const gint   value)
{
  gchar * grp;

  grp = g_strdup_printf ("Control %u", ctl);
  g_key_file_set_integer ((GKeyFile *)filter, grp, "Step", value);
  g_free (grp);
}


static const char * RGBA[N_CHANNELS] = {
  "R", "G", "B", "A"
};

gchar *
guf_filter_set_src (UserFilter *        filter,
		    const guint         channel,
		    const gchar * const value)
{
  g_key_file_set_string ((GKeyFile *)filter, "Code", RGBA[channel], value);
  return (gchar *)value;
}


gchar *
guf_filter_set_imported_src (UserFilter *        filter,
			     const guint         channel,
			     const gchar * const value)
{
  return (gchar *)guf_import_string (filter, "Code", RGBA[channel], value);
}
