/*
 * (C) 1997,2003  Jens Ch. Restemeier <jrestemeier@currantbun.com>
 *
 * (C) 2008       Torsten Neuer <tneuer@inwise.de>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *  $Id: interpreter.c,v 1.7 2008/06/16 09:41:05 tneuer Exp $
 *
 *  $Log: interpreter.c,v $
 *  Revision 1.7  2008/06/16 09:41:05  tneuer
 *  Added handlers for channel source functions (src[rgba])
 *
 *  Revision 1.6  2008/06/13 11:06:14  tneuer
 *  Added range-check optimized variants of get() / put() functions (get0 (), put0 ())
 *
 *  Revision 1.5  2008/06/11 17:52:11  tneuer
 *  Removed assertions not required anymore.
 *  Pointers to const data are const now, too.
 *
 *  Revision 1.4  2008/06/10 11:50:32  tneuer
 *  Bugfix: Comma operator inside parenthesis gives syntax error
 *
 *  Revision 1.3  2008/06/05 09:42:05  tneuer
 *  Using glib functions for directory scanning instead of <dirent.h> and <stat.h>
 *
 *  Revision 1.2  2008/05/21 12:30:37  tneuer
 *  Minor code cleanup.
 *
 *  Revision 1.1.1.1  2008/04/27 13:19:05  tneuer
 *  initial import
 *
 *  Revision 1.1  2008/04/01 16:05:31  tneuer
 *  Initial revision
 *
 */

#include "plugin-conf.h"

#include <math.h>

#include <glib/gmessages.h>

#include <libgimp/gimp.h>

#include "interpreter.h"
#include "render.h"
#include "scanner.h"
#include "parser.h"

#include "uf-pcode.h"

CodeTree * uf_code[N_CHANNELS] = { NULL, NULL, NULL, NULL };


#define CT(x) calc_tree( f->nodes[(x)] )

static gint
calc_tree (const CodeTree * const f)
{
  if (!f)
    return 0;

  switch (f->op_type)
    {
    case v_int:
      return f->value;
    case '+':
      return CT (0) + CT (1);
    case '-':
      return CT (0) - CT (1);
    case '*':
      return CT (0) * CT (1);
    case t_and:
      return CT (0) && CT (1);
    case t_or:
      return CT (0) || CT (1);
    case t_eq:
      return CT (0) == CT (1);
    case t_neq:
      return CT (0) != CT (1);
    case '<':
      return CT (0) < CT (1);
    case t_leq:
      return CT (0) <= CT (1);
    case '>':
      return CT (0) > CT (1);
    case t_geq:
      return CT (0) >= CT (1);
    case '&':
      return CT (0) & CT (1);
    case '|':
      return CT (0) | CT (1);
    case '^':
      return CT (0) ^ CT (1);
    case t_shl:
      return CT (0) << CT (1);
    case t_shr:
      return CT (0) >> CT (1);
    case '~':
      return ~CT (0);
    case 'r':
      return environment.r;
    case 'g':
      return environment.g;
    case 'b':
      return environment.b;
    case 'a':
      return environment.a;
    case 'c':
      return environment.c;
    case 'i':
      return environment.i;
    case 'u':
      return environment.u;
    case 'v':
      return environment.v;
    case 'x':
      return environment.x;
    case 'y':
      return environment.y;
    case 'z':
      return environment.z;
    case 'd':
      return environment.d;
    case 'm':
      return environment.m;
    case t_xmax:
      return environment.X;
    case t_ymax:
      return environment.Y;
    case t_zmax:
      return environment.Z;
    case t_mmax:
      return environment.M;
    case '!':
      return !CT (0);
    case t_sign:
      return -CT (0);
    case t_sin:
      return p_sin (CT (0));
    case t_cos:
      return p_cos (CT (0));
    case t_tan:
      return p_tan (CT (0));
    case t_sqr:
      return p_sqr (CT (0));
    case t_ctl:
      return p_ctl (CT (0));
    case t_get:
      return p_get (CT (0));
    case t_get0:
      return p_get0 (CT (0));
    case t_abs:
      return p_abs (CT (0));
    case t_c2d:
      return p_c2d (CT (0), CT (1));
    case t_c2m:
      return p_c2m (CT (0), CT (1));
    case t_r2x:
      return p_r2x (CT (0), CT (1));
    case t_r2y:
      return p_r2y (CT (0), CT (1));
    case '/':
      return p_div (CT (0), CT (1));
    case '%':
      return p_mod (CT (0), CT (1));
    case t_put:
      return p_put (CT (0), CT (1));
    case t_put0:
      return p_put0 (CT (0), CT (1));
    case t_map:
      return p_map (CT (0), CT (1));
    case t_add:
      return p_add (CT (0), CT (1), CT (2));
    case t_mix:
      return p_mix (CT (0), CT (1), CT (2), CT (3));
    case t_sub:
      return p_sub (CT (0), CT (1), CT (2));
    case t_min:
      return p_min (CT (0), CT (1));
    case t_max:
      return p_max (CT (0), CT (1));
    case t_rnd:
      return p_rnd (CT (0), CT (1));
    case t_dif:
      return p_dif (CT (0), CT (1));
    case t_val:
      return p_val (CT (0), CT (1), CT (2));
    case t_src:
      return p_src (CT (0), CT (1), CT (2));
    case t_srcr:
      return p_src_r (CT (0), CT (1));
    case t_srcg:
      return p_src_g (CT (0), CT (1));
    case t_srcb:
      return p_src_b (CT (0), CT (1));
    case t_srca:
      return p_src_a (CT (0), CT (1));
    case t_rad:
      return p_rad (CT (0), CT (1), CT (2));
    case '?':
      return CT (0) ? CT (1) : CT (2);
    case t_scl:
      return p_scl (CT (0), CT (1), CT (2), CT (3), CT (4));
    case t_cnv:
      return p_cnv (CT (0), CT (1), CT (2),
		    CT (3), CT (4), CT (5),
		    CT (6), CT (7), CT (8), CT (9));
    case t_comma:
      return CT (0), CT (1);
    default:
      g_message ("%s %i !\n", _("Unhandled operator"), f->op_type);
      return 0;
    }

}


static inline gint
calc (const CodeTree * const t)
{
  /*
   *  FIXME:  we do not need to clamp if the channel's formula
   *  is a do-nothing formula (environment.[crgba]) or a constant
   *  in range 0..255.
   *
   *  This could perhaps be done by adding the clamp to the
   *  code tree at compilation time.  The optimizer would then
   *  reduce in case of a constant value and the trees for the
   *  min(max(0,environment.[crgba]),255) would be trivial enough
   *  to handle, too.
   *
   *  On the other hand, we're here are machine code level and
   *  putting this into the byte-code seems a bit odd...
   */
  return p_min (p_max (0, calc_tree (t)), 255);
}


gint
R (void)
{
  return calc (uf_code[0]);
}


gint
G (void)
{
  return calc (uf_code[1]);
}


gint
B (void)
{
  return calc (uf_code[2]);
}


gint
A (void)
{
  return calc (uf_code[3]);
}
