/*
 * (C) 1997,2003  Jens Ch. Restemeier <jrestemeier@currantbun.com>
 *
 * (C) 2008       Torsten Neuer <tneuer@inwise.de>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * History:
 * 0.1 first release
 * 0.2 fixed some bugs, new gui
 * 0.3 renamed files, support for *.txt source, fixed some formulas, 
 *     deactivated PDB-interface, new preview options  
 * 0.4 new gui, reactivated PDB interface, new FF-manager, *.ffl libraries
 * 0.5 bugfixes, "About" page, some cleanup
 * 0.5a Gimp 1.1 fixes by Pedro Kaempf
 * 0.6 Gimp 1.2 fixes
 * 0.7 Patch from Artur Polaczynski to support sorting in the filter manager
 * 0.8 Compile fixes for GTK 2.0 and Gimp 1.3
 * 0.9 Gimp 2.0 fixes, bug fixes, native file format, enhanced GUI, command
 *     line tools for porting non-native filters, command line tools for
 *     creating stand-alone plugins, running filters in stand-alone mode
 *
 *
 *  $Id: main.c,v 1.22 2008/06/17 15:41:22 tneuer Exp $
 *
 *  $Log: main.c,v $
 *  Revision 1.22  2008/06/17 15:41:22  tneuer
 *  Minor code cleanups
 *
 *  Revision 1.21  2008/06/17 10:47:44  tneuer
 *  Using guf_filter_get_date_string () instead of guf_filter_get_version_string () for registering a filter
 *
 *  Revision 1.20  2008/06/16 16:01:34  tneuer
 *  Added code-only optimization to transpiler
 *
 *  Revision 1.19  2008/06/16 09:40:09  tneuer
 *  Added reduction of source function to channel source function for constant 'z' where 0 <= z <= 3
 *
 *  Revision 1.18  2008/06/11 17:52:12  tneuer
 *  Removed assertions not required anymore.
 *  Pointers to const data are const now, too.
 *
 *  Revision 1.17  2008/06/06 14:43:20  tneuer
 *  Minor optimizations in render function
 *
 *  Revision 1.16  2008/06/06 13:03:50  tneuer
 *  No more help button when running filters in stand-alone mode
 *
 *  Revision 1.15  2008/06/05 09:42:09  tneuer
 *  Using glib functions for directory scanning instead of <dirent.h> and <stat.h>
 *
 *  Revision 1.14  2008/06/02 11:25:48  tneuer
 *  Bugfix: segfault on init/run with not userfilter-path set in gimprc
 *
 *  Revision 1.13  2008/06/02 09:24:45  tneuer
 *  "make install" fixes
 *
 *  Revision 1.12  2008/05/26 15:10:56  tneuer
 *  Minor code cleanup.
 *
 *  Revision 1.11  2008/05/22 16:42:44  tneuer
 *  Added sksleton file for transpiler.
 *
 *  Revision 1.10  2008/05/21 12:30:37  tneuer
 *  Minor code cleanup.
 *
 *  Revision 1.9  2008/05/16 16:37:17  tneuer
 *  Minor code cleanup.
 *
 *  Revision 1.8  2008/05/16 14:02:54  tneuer
 *  Loading stand-alone (registered) filters.
 *
 *  Revision 1.7  2008/05/13 15:38:49  tneuer
 *  Userfilter registry run-time fixes.
 *
 *  Revision 1.6  2008/05/13 15:03:38  tneuer
 *  Using gimp_config_ functions to build paths.
 *
 *  Revision 1.5  2008/05/07 11:36:38  tneuer
 *  Registering filters for stand-alone operations.
 *
 *  Revision 1.4  2008/05/07 10:54:34  tneuer
 *  Added "userfilterrc" file for registering stand-alone filters.
 *
 *  Revision 1.3  2008/05/07 09:47:17  tneuer
 *  Ported PDB data handling to new internal filter representation.
 *
 *  Revision 1.2  2008/05/05 13:32:41  tneuer
 *  New functions guf_filter_renew_from_pdb () and guf_filter_dump () for PDB operations.
 *
 *  Revision 1.1.1.1  2008/04/27 13:19:05  tneuer
 *  initial import
 *
 *  Revision 1.1  2008/04/01 16:05:48  tneuer
 *  Initial revision
 *
 */

#include "plugin-conf.h"

#include <string.h>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

#include "main.h"
#include "checksum.h"
#include "interface.h"
#include "interpreter.h"
#include "pdb-name.h"
#include "optimizer.h"
#include "render.h"
#include "scan-filters.h"
#include "registry.h"
#include "guf.h"


typedef gboolean (*DialogFun) (GimpDrawable * drawable);


/*  Constants  */

#define PROCEDURE_NAME   "gimp_plugin_userfilter"

#define DATA_KEY_VALS    "plug_in_userfilter"
#define DATA_KEY_UI_VALS "plug_in_userfilter_ui"

#define PARASITE_KEY     "plug-in-userfilter-options"

enum {
  STANDALONE_ARGS = 5,
  USERFILTER_ARGS = 7
};


const gchar * const plugin_name     = PLUGIN_NAME;
const gchar * const userfilter_path = "userfilter-path";


/*  Local function prototypes  */

static void init (void);

static void query (void);

static void run (const gchar * const name,
		 gint                nparams,
		 const GimpParam *   param,
		 gint *              nreturn_vals,
		 GimpParam **        return_vals);


/*  Local variables  */

static gchar * data_key_vals    = NULL; /* PDB: filter save slot */
static gchar * data_key_ui_vals = NULL; /* PDB: ui save slot */
static gchar * filter_rc        = NULL; /* Path to the rc-file */


/*  Global variables */

gchar * filter_path = NULL;     /* Path to current filter */


GimpPlugInInfo PLUG_IN_INFO = {
  init,				/* init_proc  */
  NULL,				/* quit_proc  */
  query,			/* query_proc */
  run,				/* run_proc   */
};


MAIN ()


static inline gchar *
default_gimprc_entry (const gchar * const rckey,
		      gchar *             rcdefault)
{
  gchar * result;

  result = g_strdup (gimp_gimprc_query (rckey));

  if (result == NULL ||
      result[0] == '\0')
    {
      g_free (result);
      gimp_gimprc_set (rckey, rcdefault);
      result = g_strdup (gimp_gimprc_query (rckey));
    }
  g_free (rcdefault);

  return result;
}


static gchar *
default_filter_path (void)
{
  return default_gimprc_entry (userfilter_path,
			       gimp_config_build_data_path ("userfilter"));
}


static gchar *
default_filter_rc (void)
{
  GError * err = NULL;
  gchar *  path;
  gchar *  result;

  path = gimp_config_build_writable_path ("userfilterrc");
  result = gimp_config_path_expand (path, TRUE, &err);
  g_clear_error (&err);
  g_free (path);

  return result;
}


/*******************************************************************************
 *
 *  init()
 *
 *  All userfilter directories are scanned for filters that have the "Register"
 *  flag set and set them up in the gimp menu.  Path and PDB-name of the stand-
 *  alone filter are memorized in an rc-file.
 */

static inline void
userfilter_install_filter (gchar * pdb_name,
			   gchar * title,
			   gchar * description,
			   gchar * category,
			   gchar * author,
			   gchar * copyright,
			   gchar * date)
{
  static GimpParamDef args[STANDALONE_ARGS] = {
    {GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive"},
    {GIMP_PDB_IMAGE, "image", "Input image"},
    {GIMP_PDB_DRAWABLE, "drawable", "Input drawable"},
    {GIMP_PDB_INT32, "number_sliders", "Number of sliders defined (max. 8)"},
    {GIMP_PDB_INT32ARRAY, "slider_values",
     "The values for the sliders in range 0..255"},
  };
  static GimpParamDef * return_vals  = NULL;
  static int            nreturn_vals = 0;

  gimp_install_procedure (pdb_name,         /* pdb name */
			  description,      /* description */
			  description,      /* quick help */
			  author,           /* author */
			  copyright,        /* copyright */
			  date,             /* date */   
			  title, "RGB*,GRAY*",
			  GIMP_PLUGIN, 
			  G_N_ELEMENTS (args), nreturn_vals,
			  args, return_vals);

  gimp_plugin_menu_register (pdb_name, category);

  g_free (pdb_name);
  g_free (title);
  g_free (description);
  g_free (category);
  g_free (author);
  g_free (copyright);
  g_free (date);
}


static gboolean
register_filter (const UserFilter *  filter,
		 const gchar * const fpath,
		 gpointer            data )
{
  if (guf_filter_get_registered (filter))
    {
      userfilter_install_filter (registry_register_filter ((Registry *)data, filter, fpath),
				 guf_filter_get_locale_title (filter),
				 guf_filter_get_locale_description (filter),
				 guf_filter_get_category (filter),
				 guf_filter_get_author (filter),
				 guf_filter_get_copyright (filter),
				 guf_filter_get_date_string (filter));
      return TRUE;
    }
  return FALSE;
}


static void
init (void)
{
  Registry * reg;

  /*
   *  Initialize with default values
   */

  reg = registry_new ();

  filter_path = default_filter_path ();
  filter_rc = default_filter_rc ();

  /*
   *  Scan the filter directories.
   *
   *  If a filter has the registered flag set,
   *  make an entry for it in the PDB...
   */

  scan_filters (filter_path, register_filter, reg);

  registry_save (reg, filter_rc);

  /*
   *  Cleaning up.
   */

  registry_free (reg);

  g_free (filter_path);
  g_free (filter_rc);
}


/*******************************************************************************
 *
 *  query()
 */

static void
query (void)
{
  gchar * help_path;
  gchar * help_uri;

  static GimpParamDef args[USERFILTER_ARGS] = {
    {GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive"},
    {GIMP_PDB_IMAGE, "image", "Input image"},
    {GIMP_PDB_DRAWABLE, "drawable", "Input drawable"},
    {GIMP_PDB_INT32, "number_sliders", "Number of sliders defined (max. 8)"},
    {GIMP_PDB_INT32ARRAY, "slider_values",
     "The values for the sliders in range 0..255"},
    {GIMP_PDB_INT32, "number_functions",
     "The number of functions defined in order (max. 4)"},
    {GIMP_PDB_STRINGARRAY, "function", "The functions for each  channel"},
  };
  static GimpParamDef * return_vals = NULL;
  static int            nreturn_vals = 0;

  gimp_plugin_domain_register (plugin_name, LOCALEDIR);

  help_path = g_build_filename (DATADIR, "help", NULL);
  help_uri = g_filename_to_uri (help_path, NULL, NULL);
  g_free (help_path);

  gimp_plugin_help_register (plugin_name, help_uri);
  g_free (help_uri);

  gimp_install_procedure (PROCEDURE_NAME,
			  N_("Applies a user definable filter algorithm to an image."),
			  N_("A generic filter tool that uses a formula for each colour channel to modify the image."),
			  "Jens Ch. Restemeier, Torsten Neuer",
			  "(C) Jens Ch. Restemeier, Torsten Neuer",
			  "1997, 2003, 2008",
			  "_User Filter...",
			  "RGB*,GRAY*",
			  GIMP_PLUGIN, 
			  G_N_ELEMENTS (args), nreturn_vals,
			  args, return_vals);

  gimp_plugin_menu_register (PROCEDURE_NAME, "<Image>/Filters/Generic");

}


/*******************************************************************************
 *
 *  run()
 */

static GimpPDBStatusType
run_dialog (const gchar * const name,
	    gint                n_params,
	    const GimpParam *   param,
	    GimpRunMode         run_mode,
	    DialogFun           dialog,
	    GimpDrawable *      drawable)
{
  static const gint nargs[2] = {
    USERFILTER_ARGS, /* user filter mode: we can have functions as well */
    STANDALONE_ARGS  /* stand-alone mode: don't overwrite functions */
  };
  const gboolean stand_alone = (dialog == userfilter_dialog);

  switch (run_mode)
    {
    case GIMP_RUN_NONINTERACTIVE:
      if (n_params != nargs[stand_alone])
	{
	  return GIMP_PDB_CALLING_ERROR;
	}
      else
	{
	  /*
	   *  Load slider settings
	   */
	  if (param[3].data.d_int32 <= 8)
	    {
	      gint i;

	      for (i = param[3].data.d_int32; i--;)
		{
		  control.value[i] =
		    param[4].data.d_int32array[i];
		  guf_filter_set_ctl_enabled (filter, i, TRUE);
		  guf_filter_set_ctl_preset (filter, i, control.value[i]);
		}
	    }
	  if (!stand_alone)
	    {
	      /*
	       *  Load function sources...
	       */
	      if (param[5].data.d_int32 <= 4)
		{
		  gint i;

		  for (i = param[5].data.d_int32; i--;)
		    {
		      guf_filter_set_src (filter, i,
					  param[6].data.d_stringarray[i]);
		    }
		}
	    }
	}
      break;

    case GIMP_RUN_INTERACTIVE:
      /*  Possibly retrieve data  */
      {
	gint i;

	if (!stand_alone)
	  {
	    filter = guf_filter_renew_from_pdb (filter, data_key_vals);
	  }
	gimp_get_data (data_key_ui_vals, &control);

	for (i = N_CHANNELS; i--;)
	  {
	    gchar * src;
	    
	    src = guf_filter_get_src (filter, i);
	    if (!(uf_code[i] = get_code_tree (src)))
	      {
		static const gchar * const src [N_CHANNELS] = {
		  "r", "g", "b", "a"
		};
		uf_code[i] = get_code_tree (src[i]);
	      }
	    g_free (src);
	    optimize_code_tree (uf_code[i], i, FALSE, FALSE, &environment);
	  }
	
	if (!dialog (drawable))
	  {
	    return GIMP_PDB_CANCEL;
	  }
      }
      break;

    case GIMP_RUN_WITH_LAST_VALS:
      /*  Possibly retrieve data  */
      if (!stand_alone)
	{
	  filter = guf_filter_renew_from_pdb (filter, data_key_vals);
	}
      gimp_get_data (data_key_ui_vals, &control);
      break;

    default:
      break;
    }

  return GIMP_PDB_SUCCESS;
}


static void
run (const gchar * const name,
     gint                n_params,
     const GimpParam *   param,
     gint *              nreturn_vals,
     GimpParam **        return_vals)
{
  static GimpParam  values[1];
  GimpDrawable *    drawable;
  gint32            image_ID;
  GimpRunMode       run_mode;
  GimpPDBStatusType status = GIMP_PDB_CALLING_ERROR;

  *nreturn_vals = 1;
  *return_vals = values;

  bindtextdomain (GETTEXT_PACKAGE, LOCALEDIR);
#ifdef HAVE_BIND_TEXTDOMAIN_CODESET
  bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
#endif
  textdomain (GETTEXT_PACKAGE);

  run_mode = param[0].data.d_int32;
  image_ID = param[1].data.d_int32;
  drawable = gimp_drawable_get (param[2].data.d_drawable);

  /*  Initialize with default values  */

  filter = guf_filter_renew (filter);
  filter_path = default_filter_path ();
  filter_rc = default_filter_rc ();

  /*
   *  Preset the renderer environment
   */
  environment.X = gimp_drawable_width (drawable->drawable_id);
  environment.Y = gimp_drawable_height (drawable->drawable_id);
  environment.Z = gimp_drawable_bpp (drawable->drawable_id);
  environment.M = sqrt (environment.X * environment.X +
			environment.Y * environment.Y) / 2;
  environment.value = control.value;
  environment.drawable = drawable;
  environment.drawable_hash = compute_drawable_checksum (drawable);

  environment.Xm1 = environment.X - 1;
  environment.Ym1 = environment.Y - 1;

  environment.Xd2 = environment.X / 2;
  environment.Yd2 = environment.Y / 2;

  environment.rand = g_rand_new ();

  gimp_tile_cache_ntiles (drawable->ntile_cols);

  if (!strcmp (name, PROCEDURE_NAME))
    {
      /*
       *  Run User Filter IDE...
       */
      data_key_vals    = make_pdb_name (DATA_KEY_VALS, NULL, '_');
      data_key_ui_vals = make_pdb_name (DATA_KEY_UI_VALS, NULL, '_');

      status = run_dialog (name,
			   n_params, param,
			   run_mode, uf_dialog, drawable);
    }
  else if (!strncmp (name, registry_prefix, strlen (registry_prefix)))
    {
      gint i;

      /*
       *  Run in stand-alone mode...
       */
      data_key_ui_vals = make_pdb_name (DATA_KEY_UI_VALS, name, '_');
      filter = guf_filter_renew_from_registry (filter, filter_rc, name);

      for (i = N_CONTROLS; i--; )
	{
	  control.value[i] = guf_filter_get_ctl_preset (filter, i);
	}
      status = run_dialog (name,
			   n_params, param,
			   run_mode, userfilter_dialog, drawable);
    }

  if (status == GIMP_PDB_SUCCESS)
    {
      gchar * title;
      gint    i;

      for (i = 0; i < N_CHANNELS; i++)
	{
	  gchar * src;

	  src = guf_filter_get_src (filter, i);
	  if (!(uf_code[i] = get_code_tree (src)))
	    {
	      static const gchar * const src [N_CHANNELS] = {
		"r", "g", "b", "a"
	      };
	      uf_code[i] = get_code_tree (src[i]);
	    }
	  else
	    {
	      optimize_code_tree (uf_code[i], i, FALSE, TRUE, &environment);
	    }
	}
      title = guf_filter_get_locale_title (filter);
      render (drawable, NULL, (gboolean) strcmp (name, PROCEDURE_NAME), title);
      g_free (title);

      if (run_mode != GIMP_RUN_NONINTERACTIVE)
	{
	  gimp_displays_flush ();
	}
      if (run_mode == GIMP_RUN_INTERACTIVE)
	{
	  if (!strcmp (name, PROCEDURE_NAME))
	    {
	      gchar * filter_data;

	      filter_data = guf_filter_dump (filter);
	      gimp_set_data (data_key_vals, filter_data, strlen(filter_data)+1);
	      g_free (filter_data);
	    }
	  gimp_set_data (data_key_ui_vals, &control, sizeof (control));
	}
      gimp_drawable_detach (drawable);
    }

  g_free (filter_rc);
  g_free (filter_path);
  g_free (data_key_vals);
  g_free (data_key_ui_vals);

  g_rand_free (environment.rand);

  guf_filter_free (filter);

  values[0].type = GIMP_PDB_STATUS;
  values[0].data.d_status = status;
}
