/*
 * (C) 2008  Torsten Neuer <tneuer@inwise.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *  $Id: optimizer.c,v 1.7 2008/06/16 16:01:36 tneuer Exp $
 *
 *  $Log: optimizer.c,v $
 *  Revision 1.7  2008/06/16 16:01:36  tneuer
 *  Added code-only optimization to transpiler
 *
 *  Revision 1.6  2008/06/16 09:40:09  tneuer
 *  Added reduction of source function to channel source function for constant 'z' where 0 <= z <= 3
 *
 *  Revision 1.5  2008/06/13 12:44:42  tneuer
 *  Optimize get () / put () by eliminating range-checks on in-range constant storage index values
 *
 *  Revision 1.4  2008/06/10 11:50:32  tneuer
 *  Bugfix: Comma operator inside parenthesis gives syntax error
 *
 *  Revision 1.3  2008/06/05 09:42:11  tneuer
 *  Using glib functions for directory scanning instead of <dirent.h> and <stat.h>
 *
 *  Revision 1.2  2008/05/21 12:30:38  tneuer
 *  Minor code cleanup.
 *
 *  Revision 1.1.1.1  2008/04/27 13:19:05  tneuer
 *  initial import
 *
 *  Revision 1.1  2008/04/19 07:49:44  tneuer
 *  Initial revision
 *
 */

#include "plugin-conf.h"

#include <glib/gtypes.h>

#include "optimizer.h"
#include "parser.h"
#include "render.h"
#include "uf-pcode.h"


/*
 *  Simple constant subexpession elimination.
 *
 *  This might not do much for a well written filter, especially
 *  not when previewing, since the only reducing of the code tree
 *  can occur on redundant code and computing of constant values.
 *
 *  The only image dependant constants known during preview are
 *  M, X, Y, and Z, so normally not much of the tree is reduced
 *  when previewing (except when the formula is suboptimal).
 *
 *  When doing the final rendering, however, we can treat the values
 *  of the controls (and those that are directly derived from them)
 *  as constant values.  Running the optimizer here will then re-
 *  sult in an image processing that is up to 30% faster.
 *
 *  Parameters:
 *
 *  Doing "full" optimization also reduces the control values
 *  and should only be done before the final rendering.
 *
 *  The "code-only" optimization mutually excludes "full"
 *  optimization.  This is meant for (besides debugging)
 *  helping optimizing the formulas (by writing back the
 *  reduced tree as formula).
 */

static void
reduce (CodeTreePtr t,
	const gint  v)
{
  guint c;

  for (c = t->value; c--; )
    {
      g_free (t->nodes[c]);
    }
  g_free (t->nodes);
  t->op_type = v_int;
  t->value = v;
  t->nodes = NULL;
}


#define IS_INT(x) ((x)->op_type == v_int)

void
optimize_code_tree (CodeTreePtr                     t,
		    const guint                     z,
		    const gboolean                  code_only,
		    const gboolean                  full,
		    const FilterEnvironment * const e)
{
  const gboolean full_optmization = !code_only && full;

  /*
   * we can't reduce v_int further...
   */
  if (t->op_type != v_int)
    {
      /*
       *  we're inmidst the tree, so we look if we first can go further...
       */
      if (t->nodes)
	{
	  guint n;

	  for (n = t->value; n--; )
	    {
	      optimize_code_tree (t->nodes[n], z, code_only, full_optmization, e);
	    }
	  /*
	   *  We traversed down to the last branch in the tree.
	   *
	   *  If all leaves are v_int and the function is not depending
	   *  on a pixel environment, we can reduce this branch to a
	   *  leaf...
	   *
	   *  The optimizations that are most likely to happen are
	   *  those that operate on 1-2 arguments (mostly arithmetics).
	   */
	  switch (t->op_type)
	    {
	    case '!':
	      if (IS_INT (t->nodes[0]))
		{
		  reduce (t, !(t->nodes[0]->value));
		}
	      break;
	    case '~':
	      if (IS_INT (t->nodes[0]))
		{
		  reduce (t, ~(t->nodes[0]->value));
		}
	      break;
	    case t_sign:
	      if (IS_INT (t->nodes[0]))
		{
		  reduce (t, -(t->nodes[0]->value));
		}
	      break;
	    case t_abs:
	      if (IS_INT (t->nodes[0]))
		{
		  reduce (t, p_abs (t->nodes[0]->value));
		}
	      break;
	    case t_sqr:
	      if (IS_INT (t->nodes[0]))
		{
		  reduce (t, p_sqr (t->nodes[0]->value));
		}
	      break;
	    case t_sin:
	      if (IS_INT (t->nodes[0]))
		{
		  reduce (t, p_sin (t->nodes[0]->value));
		}
	      break;
	    case t_cos:
	      if (IS_INT (t->nodes[0]))
		{
		  reduce (t, p_cos (t->nodes[0]->value));
		}
	      break;
	    case t_tan:
	      if (IS_INT (t->nodes[0]))
		{
		  reduce (t, p_tan (t->nodes[0]->value));
		}
	      break;
	    case '+':
	      if ((IS_INT (t->nodes[0])) &&
		  (IS_INT (t->nodes[1])))
		{
		  reduce (t, t->nodes[0]->value + t->nodes[1]->value);
		}
	      break;
	    case '-':
	      if ((IS_INT (t->nodes[0])) &&
		  (IS_INT (t->nodes[1])))
		{
		  reduce (t, t->nodes[0]->value - t->nodes[1]->value);
		}
	      break;
	    case '*':
	      if ((IS_INT (t->nodes[0])) &&
		  (IS_INT (t->nodes[1])))
		{
		  reduce (t, t->nodes[0]->value * t->nodes[1]->value);
		}
	      break;
	    case '/':
	      /*
	       *  p_div( a, b ) {
	       *    return (b != 0) ? a / b : 1;
	       *  }
	       */
	      if (IS_INT (t->nodes[1]))
		{
		  if (t->nodes[1]->value == 0)
		    {
		      reduce (t, 1);
		    }
		  else if (IS_INT (t->nodes[0]))
		    {
		      reduce (t, p_div (t->nodes[0]->value,
					t->nodes[1]->value));
		    }
		}
	      break;
	    case '%':
	      /*
	       *  p_mod( a, b ) {
	       *    return (b != 0) ? a / b : 0;
	       *  }
	       */
	      if (IS_INT (t->nodes[1]))
		{
		  if (t->nodes[1]->value == 0)
		    {
		      reduce (t, 0);
		    }
		  else if (IS_INT (t->nodes[0]))
		    {
		      reduce (t, p_mod (t->nodes[0]->value,
					t->nodes[1]->value));
		    }
		}
	      break;
	    case '^':
	      if ((IS_INT (t->nodes[0])) &&
		  (IS_INT (t->nodes[1])))
		{
		  reduce (t, t->nodes[0]->value ^ t->nodes[1]->value);
		}
	      break;
	    case '&':
	      if ((IS_INT (t->nodes[0])) &&
		  (IS_INT (t->nodes[1])))
		{
		  reduce (t, t->nodes[0]->value & t->nodes[1]->value);
		}
	      break;
	    case '|':
	      if ((IS_INT (t->nodes[0])) &&
		  (IS_INT (t->nodes[1])))
		{
		  reduce (t, t->nodes[0]->value | t->nodes[1]->value);
		}
	      break;
	    case t_and:
	      if ((IS_INT (t->nodes[0])) &&
		  (IS_INT (t->nodes[1])))
		{
		  reduce (t, t->nodes[0]->value && t->nodes[1]->value);
		}
	      break;
	    case t_or:
	      if ((IS_INT (t->nodes[0])) &&
		  (IS_INT (t->nodes[1])))
		{
		  reduce (t, t->nodes[0]->value || t->nodes[1]->value);
		}
	      break;
	    case t_eq:
	      if ((IS_INT (t->nodes[0])) &&
		  (IS_INT (t->nodes[1])))
		{
		  reduce (t, t->nodes[0]->value == t->nodes[1]->value);
		}
	      break;
	    case t_neq:
	      if ((IS_INT (t->nodes[0])) &&
		  (IS_INT (t->nodes[1])))
		{
		  reduce (t, t->nodes[0]->value != t->nodes[1]->value);
		}
	      break;
	    case '<':
	      if ((IS_INT (t->nodes[0])) &&
		  (IS_INT (t->nodes[1])))
		{
		  reduce (t, t->nodes[0]->value < t->nodes[1]->value);
		}
	      break;
	    case '>':
	      if ((IS_INT (t->nodes[0])) &&
		  (IS_INT (t->nodes[1])))
		{
		  reduce (t, t->nodes[0]->value > t->nodes[1]->value);
		}
	      break;
	    case t_leq:
	      if ((IS_INT (t->nodes[0])) &&
		  (IS_INT (t->nodes[1])))
		{
		  reduce (t, t->nodes[0]->value <= t->nodes[1]->value);
		}
	      break;
	    case t_geq:
	      if ((IS_INT (t->nodes[0])) &&
		  (IS_INT (t->nodes[1])))
		{
		  reduce (t, t->nodes[0]->value >= t->nodes[1]->value);
		}
	      break;
	    case t_shl:
	      if ((IS_INT (t->nodes[0])) &&
		  (IS_INT (t->nodes[1])))
		{
		  reduce (t, t->nodes[0]->value << t->nodes[1]->value);
		}
	      break;
	    case t_shr:
	      if ((IS_INT (t->nodes[0])) &&
		  (IS_INT (t->nodes[1])))
		{
		  reduce (t, t->nodes[0]->value >> t->nodes[1]->value);
		}
	      break;
	    case t_dif:
	      if ((IS_INT (t->nodes[0])) &&
		  (IS_INT (t->nodes[1])))
		{
		  reduce (t, p_dif (t->nodes[0]->value,
				    t->nodes[1]->value));
		}
	      break;
	    case t_r2x:
	      if ((IS_INT (t->nodes[0])) &&
		  (IS_INT (t->nodes[1])))
		{
		  reduce (t, p_r2x (t->nodes[0]->value,
				    t->nodes[1]->value));
		}
	      break;
	    case t_r2y:
	      if ((IS_INT (t->nodes[0])) &&
		  (IS_INT (t->nodes[1])))
		{
		  reduce (t, p_r2y (t->nodes[0]->value,
				    t->nodes[1]->value));
		}
	      break;
	    case t_c2d:
	      if ((IS_INT (t->nodes[0])) &&
		  (IS_INT (t->nodes[1])))
		{
		  reduce (t, p_c2d (t->nodes[0]->value,
				    t->nodes[1]->value));
		}
	      break;
	    case t_c2m:
	      if ((IS_INT (t->nodes[0])) &&
		  (IS_INT (t->nodes[1])))
		{
		  reduce (t, p_c2m (t->nodes[0]->value,
				    t->nodes[1]->value));
		}
	      break;
	    case t_rnd:
	      /*
	       *  p_rnd( a, b ) {
	       *    return (a != b) ? (rand() % (b - a)) + a : a;
	       *  }
	       *
	       *  we can only optimize the case where (a == b)...
	       */
	      if ((IS_INT (t->nodes[0])) &&
		  (IS_INT (t->nodes[1])) &&
		  (t->nodes[0]->value == t->nodes[1]->value))
		{
		  reduce (t, t->nodes[0]->value);
		}
	      break;
	    case t_min:
	      if ((IS_INT (t->nodes[0])) &&
		  (IS_INT (t->nodes[1])))
		{
		  reduce (t, p_min (t->nodes[0]->value,
				    t->nodes[1]->value));
		}
	      break;
	    case t_max:
	      if ((IS_INT (t->nodes[0])) &&
		  (IS_INT (t->nodes[1])))
		{
		  reduce (t, p_max (t->nodes[0]->value,
				    t->nodes[1]->value));
		}
	      break;
	    case t_add:
	      if ((IS_INT (t->nodes[0])) &&	/* a */
		  (IS_INT (t->nodes[1])))	/* b */
		{
		  if (IS_INT (t->nodes[2]))	/* c */
		    {
		      reduce (t, p_add (t->nodes[0]->value,
					t->nodes[1]->value,
					t->nodes[2]->value));
		    }
		  else
		    {
		      /*
		       *    p_add( a, b, c ) {
		       *      return p_min( a+b, c );
		       *    }
		       */
		      t->op_type = t_min;
		      t->value = 2;
		      t->nodes[0]->value += t->nodes[1]->value;
		      g_free (t->nodes[1]);
		      t->nodes[1] = t->nodes[2];
		      t->nodes[2] = NULL;
		    }
		}
	      break;
	    case t_sub:
	      if ((IS_INT (t->nodes[0])) &&
		  (IS_INT (t->nodes[1])))
		{
		  if (IS_INT (t->nodes[2]))
		    {
		      reduce (t, p_sub (t->nodes[0]->value,
					t->nodes[1]->value,
					t->nodes[2]->value));
		    }
		  else
		    {
		      /*
		       *    p_sub( a, b, c ) {
		       *      return p_max( p_dif(a,b), c );
		       *    }
		       */
		      t->op_type = t_max;
		      t->value = 2;
		      t->nodes[0]->value = p_dif (t->nodes[0]->value,
						  t->nodes[1]->value);
		      g_free (t->nodes[1]);
		      t->nodes[1] = t->nodes[2];
		      t->nodes[2] = NULL;
		    }
		}
	      break;
	    case t_scl:
	      if ((IS_INT (t->nodes[1])) &&	/* il */
		  (IS_INT (t->nodes[2])))	/* ih */
		{
		  /*
		   *  p_scl(a,il,ih,ol,oh) {
		   *    return (ih!=il) ? (((a-il)*(oh-ol))/(ih-il))+ol : 0;
		   *  }
		   */
		  if (t->nodes[1]->value == t->nodes[2]->value)
		    {
		      reduce (t, 0);
		    }
		  else if ((IS_INT (t->nodes[0])) &&	/* a */
			   (IS_INT (t->nodes[3])) &&	/* ol */
			   (IS_INT (t->nodes[4])))	/* oh */
		    {
		      reduce (t, p_scl (t->nodes[0]->value,
					t->nodes[1]->value,
					t->nodes[2]->value,
					t->nodes[3]->value,
					t->nodes[4]->value));
		    }
		}
	      break;
	    case t_comma:
	      if (IS_INT (t->nodes[0]))
		{
		  CodeTree **n = t->nodes;

		  /*
		   * value left of comma is redundant
		   */
		  *t = *n[1];
		  g_free (n[0]);
		  g_free (n);
		}
	      break;
	    case '?':
	      if (IS_INT (t->nodes[0]))
		{
		  CodeTree **n = t->nodes;

		  /*
		   *  If the boolean expression of the '?:' statement is a constant,
		   *  we can dump the unreached code and reduce the tree.
		   */
		  if (n[0]->value != 0)
		    {
		      *t = *n[1];
		      g_free (n[1]);
		      free_code_tree (n[2]);
		    }
		  else
		    {
		      *t = *n[2];
		      g_free (n[2]);
		      free_code_tree (n[1]);
		    }
		  g_free (n[0]);
		  g_free (n);
		}
	      break;
	      /*
	       *  Optimization of "src" function.
	       *  This can be done for constant 'z', where 0 <= 'z' <= 3.
	       */
	    case t_src:
	      if (IS_INT (t->nodes[2]))
		{
		  switch (t->nodes[2]->value)
		    {
		    case 0:
		      t->op_type = t_srcr;
		      break;
		    case 1:
		      t->op_type = t_srcg;
		      break;
		    case 2:
		      t->op_type = t_srcb;
		      break;
		    case 3:
		      t->op_type = t_srca;
		      break;
		    default:
		      break;
		    }
		}
	      break;
	      /*
	       *  Control-dependant functions may only be optimized when
	       *  "full optmization" is requested, i.e. during the final
	       *  rendering...
	       */
	    case t_ctl:
	      if (full_optmization &&
		  (IS_INT (t->nodes[0])))
		{
		  reduce (t, p_ctl (t->nodes[0]->value));
		}
	      break;
	    case t_map:
	      if (full_optmization &&
		  (IS_INT (t->nodes[0])) &&
		  (IS_INT (t->nodes[1])))
		{
		  reduce (t, p_map (t->nodes[0]->value,
				    t->nodes[1]->value));
		}
	      break;
	    case t_val:
	      if (full_optmization &&
		  (IS_INT (t->nodes[0])) &&
		  (IS_INT (t->nodes[1])) &&
		  (IS_INT (t->nodes[2])))
		{
		  reduce (t, p_val (t->nodes[0]->value,
				    t->nodes[1]->value,
				    t->nodes[2]->value));
		}
	      break;
	      /*
	       *  Functions related to the temporary storage (get () / put ())
	       *  cannot normally be optimized by reducing to a constant value.
	       *  We can, however, switch to the non-range-checking function
	       *  if the storage index argument is a constant and in range of
	       *  the storage array.  We can also reduce them as usual if the
	       *  index (if a constant value) is out of range (which is rarely
	       *  the case)...
	       */
	    case t_get:
	      if (IS_INT (t->nodes[0]))
		{
		  if ((t->nodes[0]->value >= 0) && (t->nodes[0]->value < N_TEMPSLOTS))
		    {
		      t->op_type = t_get0;
		    }
		  else
		    {
		      reduce (t, 0);
		    }
		}
	      break;
	    case t_put:
	      if (IS_INT (t->nodes[1]))
		{
		  if ((t->nodes[0]->value >= 0) && (t->nodes[0]->value < N_TEMPSLOTS))
		    {
		      t->op_type = t_put0;
		    }
		  else
		    {
		      reduce (t, 0);
		    }
		}
	      break;
	    default:
	      break;
	    }
	}
      else
	{
	  /*
	   *  We reached a leaf in the tree.
	   *
	   *  Change image dependant constants into v_int operations.
	   *
	   *  Note:
	   *
	   *  These values could also have been set by the parser to
	   *  v_int - they haven't been however, as the parser is also
	   *  used by the transpiler uf2c(1) which does not now about
	   *  the image environment at compile time...
	   */
	  switch (t->op_type)
	    {
	    case 'z':
	      reduce (t, z);
	      break;
	    default:
	      break;
	    }
	  if (!code_only)
	    {
	      g_assert (e != NULL);
	      switch (t->op_type)
		{
		case t_mmax:
		  t->op_type = v_int;
		  t->value = e->M;
		  break;
		case t_xmax:
		  t->op_type = v_int;
		  t->value = e->X;
		  break;
		case t_ymax:
		  t->op_type = v_int;
		  t->value = e->Y;
		  break;
		case t_zmax:
		  t->op_type = v_int;
		  t->value = e->Z;
		  break;
		default:
		  break;
		}
	    }
	}
    }
}
