/*
 * (C) 2008  Torsten Neuer <tneuer@inwise.de>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *  $Id: registry.c,v 1.8 2008/06/17 15:41:27 tneuer Exp $
 *
 *  $Log: registry.c,v $
 *  Revision 1.8  2008/06/17 15:41:27  tneuer
 *  Minor code cleanups
 *
 *  Revision 1.7  2008/06/11 17:52:22  tneuer
 *  Removed assertions not required anymore.
 *  Pointers to const data are const now, too.
 *
 *  Revision 1.6  2008/06/05 15:23:55  tneuer
 *  Use g_fopen () for UTF-8 compatibility
 *
 *  Revision 1.5  2008/05/16 14:02:55  tneuer
 *  Loading stand-alone (registered) filters.
 *
 *  Revision 1.4  2008/05/13 15:38:49  tneuer
 *  Userfilter registry run-time fixes.
 *
 *  Revision 1.3  2008/05/13 15:03:38  tneuer
 *  Using gimp_config_ functions to build paths.
 *
 *  Revision 1.2  2008/05/07 11:36:38  tneuer
 *  Registering filters for stand-alone operations.
 *
 *  Revision 1.1  2008/05/07 10:54:34  tneuer
 *  Added "userfilterrc" file for registering stand-alone filters.
 *
 */

#include "plugin-conf.h"

#include <glib/gstdio.h>

#include "registry.h"
#include "pdb-name.h"


const gchar * const registry_prefix = "user_filter";


Registry *
registry_renew (Registry * registry)
{
  if (registry)
    {
      registry_free (registry);
    }
  return registry_new ();
}


Registry *
registry_load (Registry *          registry,
	       const gchar * const filename)
{
  GError * err = NULL;

  registry = registry_renew (registry);
  g_key_file_load_from_file ((GKeyFile *)registry, filename, G_KEY_FILE_NONE, &err);
  if (err)
    {
      g_warning ("%s '%s': %s\n",
		 _("Error loading registry"), filename, err->message);
      g_error_free (err);
    }
  return registry;
}


gboolean
registry_save (const Registry *    registry,
	       const gchar * const filename)
{
  GError * err = NULL;
  gchar *  buf;
  gsize    len;
  FILE *   f;

  buf = g_key_file_to_data ((GKeyFile *)registry, &len, &err);

  if (err)
    {
      g_warning ("%s '%s': %s\n",
		 _("Error saving registry"), filename, err->message);
      g_error_free (err);
      g_free (buf);
      return FALSE;
    }

  if ((f = g_fopen(filename,"w")) == NULL)
    {
      g_warning ("%s '%s': %s\n",
		 _("Error saving registry"), filename, _("File open error"));
      g_free (buf);
      return FALSE;
    }

  fputs (buf, f);
  fclose (f);
  g_free (buf);

  return TRUE;
}


gchar *
registry_register_filter (Registry *          registry,
			  const UserFilter *  filter,
			  const gchar * const fpath)
{
  gchar * title;
  gchar * key;

  title = guf_filter_get_title (filter);
  key = make_pdb_name (registry_prefix, title, '_');
  g_key_file_set_string ((GKeyFile *)registry, "UserFilter", key, fpath);
  g_free (title);

  return key;
}


gchar *
registry_get_filter_path (const Registry *    registry,
			  const gchar * const entry)
{
  GError * err = NULL;
  gchar *  val;

  val = g_key_file_get_string ((GKeyFile *)registry, "UserFilter", entry, &err);
  if (err != NULL)
    {
      g_warning ("registry_get_filter_path (): %s: '%s'", err->message, entry);
      g_error_free (err);
      g_free (val);
      val = NULL;
    }
  return val;
}
