/*
 * (C) 1997,2003  Jens Ch. Restemeier <jrestemeier@currantbun.com>
 *
 * (C) 2008       Torsten Neuer <tneuer@inwise.de>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *  $Id: render.c,v 1.11 2008/06/16 09:38:21 tneuer Exp $
 *
 *  $Log: render.c,v $
 *  Revision 1.11  2008/06/16 09:38:21  tneuer
 *  Bugfix: 'z' parameter in source function must be unsigned
 *  Added channel variants of source function
 *
 *  Revision 1.10  2008/06/13 12:45:49  tneuer
 *  Add support for different edge modes for the pixel fetcher
 *
 *  Revision 1.9  2008/06/12 14:47:31  tneuer
 *  Using gimp_progress_init_printf ()
 *
 *  Revision 1.8  2008/06/11 17:52:28  tneuer
 *  Removed assertions not required anymore.
 *  Pointers to const data are const now, too.
 *
 *  Revision 1.7  2008/06/07 14:58:25  tneuer
 *  New cosine and tangens tables to increase accuracy
 *
 *  Revision 1.6  2008/06/06 14:43:24  tneuer
 *  Minor optimizations in render function
 *
 *  Revision 1.5  2008/05/29 11:46:17  tneuer
 *  Minor code cleanup.
 *
 *  Revision 1.4  2008/05/27 15:09:23  tneuer
 *  Bugfix: development interface overwrites previous control values on re-run
 *
 *  Revision 1.3  2008/05/22 11:03:44  tneuer
 *  New callback functions for source and meta-info entry widgets.
 *
 *  Revision 1.2  2008/05/21 12:30:41  tneuer
 *  Minor code cleanup.
 *
 *  Revision 1.1.1.1  2008/04/27 13:19:02  tneuer
 *  initial import
 *
 *  Revision 1.1  2008/04/01 16:06:24  tneuer
 *  Initial revision
 *
 */

#include "plugin-conf.h"

#include <stdlib.h>

#include <gtk/gtk.h>

#include <libgimp/gimp.h>

#include "main.h"
#include "render.h"
#include "interpreter.h"
#include "uf-pcode.h"


static GimpPixelFetcher * fetcher;
static guchar             pixbuf[N_CHANNELS];


/*
 *  Retrieve pixel information for (xx,yy) in channel (zz)
 *
 *  FIXME:  Using a static pixbuf makes the plug-in thread-unsafe.
 *          However, since plug-ins are not run in multi-threaded mode, this
 *          is a minor speed-up...
 */

gint drawable_src_r (const gint xx,
		     const gint yy)
{
  gimp_pixel_fetcher_get_pixel (fetcher, xx, yy, pixbuf);
  return pixbuf[0];
}


gint drawable_src_g (const gint xx,
		     const gint yy)
{
  gimp_pixel_fetcher_get_pixel (fetcher, xx, yy, pixbuf);
  return pixbuf[1];
}


gint drawable_src_b (const gint xx,
		     const gint yy)
{
  gimp_pixel_fetcher_get_pixel (fetcher, xx, yy, pixbuf);
  return pixbuf[2];
}


gint drawable_src_a (const gint xx,
		     const gint yy)
{
  gimp_pixel_fetcher_get_pixel (fetcher, xx, yy, pixbuf);
  return pixbuf[3];
}


gint
drawable_src (const gint  xx,
	      const gint  yy,
	      const guint zz)
{
  if (zz < environment.Z)
    {
      gimp_pixel_fetcher_get_pixel (fetcher, xx, yy, pixbuf);
      return pixbuf[zz];
    }
  return 255;
}


static inline void
render_func_prepare (const gint     x,
		     const gint     y,
		     const gboolean gray,
		     GimpPreview *  preview)
{
  if (preview)
    {
      gimp_preview_untransform (preview,
				x, y, &environment.x, &environment.y);
    }
  else
    {
      environment.x = x;
      environment.y = y;
    }
  environment.d = p_c2d (environment.Xd2 - environment.x,
			 environment.Yd2 - environment.y);
  environment.m = p_c2m (environment.Xd2 - environment.x,
			 environment.Yd2 - environment.y);
  environment.r = p_src (environment.x, environment.y, 0);
  if (gray)
    {
      environment.g =
	environment.b = environment.r;
      environment.a = p_src (environment.x, environment.y, 1);
    }
  else
    {
      environment.g = p_src (environment.x, environment.y, 1);
      environment.b = p_src (environment.x, environment.y, 2);
      environment.a = p_src (environment.x, environment.y, 3);
    }
  environment.i = ((76 * environment.r) +
		   (150 * environment.g) + (29 * environment.b)) / 256;
  environment.u = ((-19 * environment.r) +
		   (-37 * environment.g) + (56 * environment.b)) / 256;
  environment.v = ((78 * environment.r) +
		   (-65 * environment.g) + (-13 * environment.b)) / 256;
}


static void
render_func_gray (const gint x,
		  const gint y,
		  guchar *   dst,
		  const gint bpp,
		  gpointer   data)
{
  render_func_prepare (x, y, TRUE, GIMP_PREVIEW (data));
  environment.z = 0;
  environment.c = (environment.r + environment.g + environment.b) / 3;
  dst[0] = (R () + G () + B ()) / 3;
  if (bpp > 1)
    {
      environment.z = 1;
      environment.c = environment.a;
      dst[1] = A ();
    }
}


typedef int (* ChannelFunc) (void);

static void
render_func_rgb (const gint x,
		 const gint y,
		 guchar *   dst,
		 const gint bpp,
		 gpointer   data)
{
  static const gint * rgba [N_CHANNELS] = {
    &environment.r,
    &environment.g,
    &environment.b,
    &environment.a
  };
  static const ChannelFunc C[N_CHANNELS] = {
    R, G, B, A
  };
  gint z;

  render_func_prepare (x, y, FALSE, GIMP_PREVIEW (data));
  for (z = 0; z < bpp; z++)
    {
      environment.z = z;
      environment.c = *rgba[z];
      dst[z] = C[z] ();
    }
}


/*
 *  FIXME:
 *
 *  To make the function thread-safe, a separate copy of the pixel environment
 *  must exist per thread.  Also, a filter using the rnd() function may not be
 *  multi-threaded at all as this breaks the random number sequence.
 */

void
render (GimpDrawable *       drawable,
	GimpPreview *        preview,
	const gboolean       stand_alone,
	const gchar *  const title)
{
  static const GimpRgnFuncDest render_func[2] = {
    render_func_gray,
    render_func_rgb
  };
  GimpRGB bgcolor;
  const gboolean img_rgb = gimp_drawable_is_rgb (drawable->drawable_id);

  g_rand_set_seed (environment.rand, environment.drawable_hash);
  fetcher = gimp_pixel_fetcher_new (drawable, FALSE);
  if (!gimp_context_get_background (&bgcolor))
    {
      gimp_rgb_set_uchar (&bgcolor, 255, 255, 255);
    }
  gimp_pixel_fetcher_set_bg_color (fetcher, &bgcolor);
  gimp_pixel_fetcher_set_edge_mode (fetcher, control.edge_mode);
  if (preview)
    {
      guchar * buf;
      guchar * p;
      gint     x;
      gint     y;
      gint     w;
      gint     h;
      gint     bpp;

      buf = gimp_zoom_preview_get_source (GIMP_ZOOM_PREVIEW (preview),
					  &w, &h, &bpp);
      for (p = buf, y = 0; y < h; y++)
	{
	  for (x = 0; x < w; x++)
	    {
	      render_func[img_rgb] (x, y, p, bpp, preview);
	      p += bpp;
	    }
	}
      gimp_preview_draw_buffer (preview, buf, w * bpp);
      g_free (buf);
    }
  else
    {
      GimpRgnIterator * iterator;

      if (title != NULL)
	{
	  if (stand_alone)
	    {
	      gimp_progress_init_printf ("%s %s", _("Applying"), title);
	    }
	  else
	    {
	      gimp_progress_init_printf ("%s: '%s'", _("Applying User Filter"), title);
	    }
	}
      else
	{
	  gimp_progress_init (_("Applying User Filter..."));
	}
      iterator = gimp_rgn_iterator_new (drawable, 0);
      gimp_rgn_iterator_dest (iterator, render_func[img_rgb], NULL);
      gimp_rgn_iterator_free (iterator);
      gimp_progress_end ();
    }
  gimp_pixel_fetcher_destroy (fetcher);
}
