/*
 * (C) 1997,2003  Jens Ch. Restemeier <jrestemeier@currantbun.com>
 *
 * (C) 2008       Torsten Neuer <tneuer@inwise.de>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *  $Id: scan-filters.c,v 1.6 2008/06/11 17:52:30 tneuer Exp $
 *
 *  $Log: scan-filters.c,v $
 *  Revision 1.6  2008/06/11 17:52:30  tneuer
 *  Removed assertions not required anymore.
 *  Pointers to const data are const now, too.
 *
 *  Revision 1.5  2008/06/07 17:02:24  tneuer
 *  Using g_ascii_strcasecmp () instead of strcasecmp ()
 *
 *  Revision 1.4  2008/06/05 09:42:13  tneuer
 *  Using glib functions for directory scanning instead of <dirent.h> and <stat.h>
 *
 *  Revision 1.3  2008/05/26 15:10:56  tneuer
 *  Minor code cleanup.
 *
 *  Revision 1.2  2008/05/21 12:30:41  tneuer
 *  Minor code cleanup.
 *
 *  Revision 1.1.1.1  2008/04/27 13:19:05  tneuer
 *  initial import
 *
 *  Revision 1.1  2008/04/23 15:06:16  tneuer
 *  Initial revision
 *
 */

#include "plugin-conf.h"

#include <string.h>

#include <glib.h>
#include <glib/gstdio.h>

#include <libgimp/gimp.h>

#include "scan-filters.h"
#include "guf.h"


static inline gboolean
is_filter_file (const gchar * const s)
{
  const gchar * x = strrchr(s, '.');

  return ( (x != NULL) && !g_ascii_strcasecmp (x, EXT_GUF) );
}


static inline gchar *
fix_home_dir (gchar * path)
{
  gchar * tmp;

  tmp = g_strdup_printf ("%s%s", g_get_home_dir (), path + 1);
  g_free (path);
  return tmp;
}


static void
scan_filter_dir (gchar *             path, 
		 const FilterHandler handler,
		 gpointer            data)
{
  GError * err = NULL;
  GDir *   dir;

  if (!path || !path[0])
    {
      return;
    }

  /* correct path for home directory */
  if (path[0] == '~')
    {
      path = fix_home_dir (path);
    }

  /* cut off trailing slashes */
  if (G_IS_DIR_SEPARATOR (path[strlen (path)]))
    {
      path[strlen (path)] = '\0';
    }

  if ((dir = g_dir_open (path, 0, &err)) != NULL)
    {
      const gchar * entry;

      while ((entry = g_dir_read_name (dir)) != NULL)
	{
	  struct stat filestat;
	  gchar *     fname;

	  fname = g_strdup_printf ("%s%c%s",
				   path, G_DIR_SEPARATOR, entry);
	  if (!g_stat (fname, &filestat))
	    {
	      if (S_ISDIR (filestat.st_mode))
		{
		  scan_filter_dir (fname, handler, data);
		}
	      else if (S_ISREG (filestat.st_mode) &&
		       is_filter_file (fname))
		{
		  UserFilter * filter;
		  
		  filter = guf_filter_new ();
		  if (guf_read (fname, GUF_PROTOCOL, filter))
		    {
		      handler (filter, fname, data);
		    }
		  else
		    {
		      g_warning ("%s '%s' !\n", _("Error loading"), fname);
		    }
		  guf_filter_free (filter);
		}
	    }
	  g_free (fname);
	}
      g_dir_close (dir);
    }
  else
    {
      if (err != NULL)
	{
	  g_warning ("%s: %s", _("Unable to scan directory"), err->message);
	  g_error_free (err);
	}
    }
}


void
scan_filters (const gchar * const path, 
	      const FilterHandler handler,
	      gpointer            data)
{
  GList * dir;
  GList * start;

  start = gimp_path_parse (path, 99, TRUE, NULL);
  for (dir = start; (dir != NULL); dir = dir->next)
    {
      scan_filter_dir (dir->data, handler, data);
    }
  gimp_path_free (start);
}
