/*
 * (C) 2008  Torsten Neuer <tneuer@inwise.de>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *  $Id: txt2guf.c,v 1.8 2008/06/20 10:15:21 tneuer Exp $
 *
 *  $Log: txt2guf.c,v $
 *  Revision 1.8  2008/06/20 10:15:21  tneuer
 *  Added output-directory option
 *
 *  Revision 1.7  2008/06/11 17:52:41  tneuer
 *  Removed assertions not required anymore.
 *  Pointers to const data are const now, too.
 *
 *  Revision 1.6  2008/06/07 17:21:28  tneuer
 *  Using g_fprintf () instead of fprintf () to avoid UTF-8 pitfalls
 *
 *  Revision 1.5  2008/05/29 11:46:18  tneuer
 *  Minor code cleanup.
 *
 *  Revision 1.4  2008/05/27 10:55:43  tneuer
 *  Added man-pages for command-line utility programs.
 *
 *  Revision 1.3  2008/05/26 15:10:56  tneuer
 *  Minor code cleanup.
 *
 *  Revision 1.2  2008/05/23 10:08:46  tneuer
 *  Added command-line filter conversion utilities to compile set
 *
 *  Revision 1.1  2008/05/22 19:06:07  tneuer
 *  Initial revision.
 *
 */

#include "plugin-conf.h"

#include <stdio.h>
#include <getopt.h>
#include <string.h>
#include <libgen.h>
#include <stdlib.h>

#include <glib.h>
#include <glib/gprintf.h>

#include "txt.h"
#include "guf.h"

static guint verbose = 0;
static guint count   = 0;


static gboolean
compile (const gchar * const src,
	 const gchar * const dst)
{
  UserFilter * filter;

  if (dst != NULL)
    {
      if (g_mkdir_with_parents (dst, 0750) < 0)
	{
	  g_fprintf (stderr, "%s: '%s'\n",
		     _("Failed to create output directory"), dst);
	  return FALSE;
	}
    }

  filter = guf_filter_new ();
  if (verbose)
    {
      g_fprintf (stderr, "%s '%s' ...\n", 
		 _("Processing"), src);
    }
  if (!import_txt (src, filter))
    {
      guf_filter_free (filter);
      return FALSE;
    }
  else
    {
      gchar * fn;
      gchar * x;
      gchar * t;

      if (dst != NULL)
	{
	  fn = g_strdup_printf ("%s" G_DIR_SEPARATOR_S "%s", dst, src);
	}
      else
	{
	  fn = g_strdup (src);
	}

      if ( ((x = strrchr (fn, '.')) != NULL) &&
	   (strlen (x) == 4) )
	{
	  strcpy (x,EXT_GUF);
	}

      if (verbose)
	{
	  g_fprintf (stderr, "+ %s '%s' -> '%s' ...\n",
		     _("converting"), src, fn);
	  count ++;
	}

      t = g_path_get_basename (src);
      x = g_strdup_printf (" %s '%s' %s %s v" PACKAGE_VERSION ".",
			   _("Converted from"), t, _("with"), g_get_prgname());
      guf_save (fn, x, GUF_LEVEL_1, filter);

      g_free (fn);
      g_free (t);
      g_free (x);
    }
  guf_filter_free (filter);
  return TRUE;
}


static void
print_help (void)
{
  g_fprintf (stderr,
	     _("\n"
	       "Usage: %s [option] <filter%s> ...\n"
	       "\n"
	       "Convert Plugin Commander source files (%s) to User Filter (%s) files.\n"
	       "\n"
	       "Options:\n"
	       "  -h | --help :\n"
	       "    Display this help.\n"
	       "  -o | --output-directory name :\n"
	       "    Write exported filters to the specified directory.\n"
	       "    The directory will be created if it does not exist.\n"
	       "  -v | --verbose :\n"
	       "    Increase verbosity level.\n"
	       "  -V | --version :\n"
	       "    Display version information.\n" "\n"),
	     g_get_prgname (), EXT_TXT, EXT_TXT, EXT_GUF);
}


static inline void
set_prg_name (const gchar * const path)
{
  gchar * t;

  t = g_path_get_basename (path);
  g_set_prgname (t);
  g_free (t);
}


gint
main (gint     argc,
      gchar ** argv)
{
  gchar * outdir = NULL;

  bindtextdomain (GETTEXT_PACKAGE, LOCALEDIR);
#ifdef HAVE_BIND_TEXTDOMAIN_CODESET
  bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
#endif
  textdomain (GETTEXT_PACKAGE);

  set_prg_name (argv[0]);

  while (TRUE)
    {
      static struct option long_options[] = {
	/* name               arg, val, short */
	{"help",              FALSE, NULL, 'h'},
	{"output-directory",  TRUE,  NULL, 'o'},
	{"verbose",           FALSE, NULL, 'v'},
	{"version",           FALSE, NULL, 'V'},
	{NULL,                FALSE, NULL, '\0'}
      };
      gint option_index = 0;
      gint c;

      if ((c = getopt_long (argc, argv, "ho:vV",
			    long_options, &option_index)) < 0)
	{
	  break;
	}

      switch (c)
	{
	case '?':
	case 'h':
	  print_help ();
	  return ((c == 'h') ? EXIT_SUCCESS : EXIT_FAILURE);
	case 'o':
	  outdir = optarg;
	  break;
	case 'v':
	  ++verbose;
	  break;
	case 'V':
	  g_fprintf (stderr, "%s v" PACKAGE_VERSION "\n", g_get_prgname ());
	  return EXIT_SUCCESS;
	}
    }

  if (optind >= argc)
    {
      print_help ();
      return EXIT_FAILURE;
    }
  else
    {
      while (optind < argc)
	{
	  if (!compile (argv[optind++], outdir))
	    {
	      return EXIT_FAILURE;
	    }
	}
    }

  if (verbose)
    {
      g_fprintf (stderr, _("Done with %u filters.\n"), count);
    }
  
  return EXIT_SUCCESS;
}
