/*
 * (C) 1997  Jens Ch. Restemeier <jchrr@hrz.uni-bielefeld.de>
 *
 * (C) 2008  Torsten Neuer <tneuer@inwise.de>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *  $Id: uf2c.c,v 1.20 2008/06/17 09:59:43 tneuer Exp $
 *
 *  $Log: uf2c.c,v $
 *  Revision 1.20  2008/06/17 09:59:43  tneuer
 *  Added verbose messages for code stripping
 *
 *  Revision 1.19  2008/06/17 09:54:38  tneuer
 *  Added option '-S' ('--strip') to strip filter dump (remove comments & code)
 *
 *  Revision 1.18  2008/06/17 09:08:53  tneuer
 *  Bugfix: Missing newline before #line directive
 *
 *  Revision 1.17  2008/06/17 08:45:31  tneuer
 *  Intermediate code optimization is optional now.
 *  Optimization results printed in verbose mode.
 *
 *  Revision 1.16  2008/06/16 16:01:39  tneuer
 *  Added code-only optimization to transpiler
 *
 *  Revision 1.15  2008/06/11 17:52:47  tneuer
 *  Removed assertions not required anymore.
 *  Pointers to const data are const now, too.
 *
 *  Revision 1.14  2008/06/10 11:50:37  tneuer
 *  Bugfix: Comma operator inside parenthesis gives syntax error
 *
 *  Revision 1.13  2008/06/07 17:21:28  tneuer
 *  Using g_fprintf () instead of fprintf () to avoid UTF-8 pitfalls
 *
 *  Revision 1.12  2008/06/06 14:06:27  tneuer
 *  Compiled filters now should have correct locale set
 *
 *  Revision 1.11  2008/06/05 15:24:16  tneuer
 *  Use g_fopen () for UTF-8 compatibility
 *
 *  Revision 1.10  2008/06/04 10:09:40  tneuer
 *  *** empty log message ***
 *
 *  Revision 1.9  2008/06/02 09:24:45  tneuer
 *  "make install" fixes
 *
 *  Revision 1.8  2008/05/29 11:45:36  tneuer
 *  Default location for skeleton file is now in the standard data directory for this application
 *
 *  Revision 1.7  2008/05/27 10:55:43  tneuer
 *  Added man-pages for command-line utility programs.
 *
 *  Revision 1.6  2008/05/26 15:10:23  tneuer
 *  Using g_get_prgname () now to fetch the program name.
 *
 *  Revision 1.5  2008/05/23 17:47:07  tneuer
 *  Pragma #line fixes in uf2c / plug-in skeleton file
 *
 *  Revision 1.4  2008/05/23 15:26:20  tneuer
 *  uf2c now producing working native plug-ins from filter files
 *
 *  Revision 1.3  2008/05/22 16:42:44  tneuer
 *  Added skeleton file for transpiler.
 *
 *  Revision 1.2  2008/05/21 12:30:49  tneuer
 *  Minor code cleanup.
 *
 *  Revision 1.1.1.1  2008/04/27 13:19:05  tneuer
 *  initial import
 *
 *  Revision 1.1  2008/04/02 16:26:43  tneuer
 *  Initial revision
 *
 */

#include "plugin-conf.h"

#include <stdio.h>
#include <getopt.h>
#include <string.h>
#include <stdlib.h>

#include <glib/gtypes.h>
#include <glib/gstdio.h>
#include <libgimp/gimp.h>

#include "parser.h"
#include "scanner.h"
#include "pdb-name.h"
#include "optimizer.h"
#include "userfilter.h"
#include "guf.h"

#define DEFAULT_SKELETON DATADIR G_DIR_SEPARATOR_S "userfilter.skl"

static guint    verbose       = 0;                /* level of verbosity */
static guint    line          = 1;                /* line count */
static guint    sline         = 1;                /* skeleton line count */
static gchar *  skeleton_name = DEFAULT_SKELETON; /* default skeleton file */
static gchar *  outfile_name  = NULL;             /* default output file (stdout) */
static gboolean optimize      = FALSE;            /* code-optimize the formulas */
static gboolean strip         = FALSE;            /* strip comments and code from filter dump */

FilterEnvironment environment; /* dummy environment / required for optimizer */


static guint
count_nodes (const CodeTree * const t)
{
  if (t != NULL)
    {
      guint n = 1;

      if (t->nodes != NULL)
	{
	  gint  i;

	  for (i=0; i < t->value; i++)
	    {
	      n += count_nodes (t->nodes[i]);
	    }
	}
      return n;
    }
  return 0;
}


#define CT(x) calc_tree(f, a->nodes[(x)])

static void
calc_tree (FILE *                 f,
	   const CodeTree * const a)
{
  if (f)
    {
      fputc ('(', f);
      switch (a->op_type)
	{
	case v_int:
	  g_fprintf (f, "%d", a->value);
	  break;
	case '+':
	case '-':
	case '*':
	case '<':
	case '>':
	case '&':
	case '|':
	case '^':
	  CT (0);
	  fputc (a->op_type, f);
	  CT (1);
	  break;
	case t_and:
	  CT (0);
	  g_fprintf (f, "&&");
	  CT (1);
	  break;
	case t_or:
	  CT (0);
	  g_fprintf (f, "||");
	  CT (1);
	  break;
	case t_eq:
	  CT (0);
	  g_fprintf (f, "==");
	  CT (1);
	  break;
	case t_neq:
	  CT (0);
	  g_fprintf (f, "!=");
	  CT (1);
	  break;
	case t_leq:
	  CT (0);
	  g_fprintf (f, "<=");
	  CT (1);
	  break;
	case t_geq:
	  CT (0);
	  g_fprintf (f, ">=");
	  CT (1);
	  break;
	case t_shl:
	  CT (0);
	  g_fprintf (f, "<<");
	  CT (1);
	  break;
	case t_shr:
	  CT (0);
	  g_fprintf (f, ">>");
	  CT (1);
	  break;
	case 'r':
	case 'g':
	case 'b':
	case 'a':
	case 'c':
	case 'i':
	case 'u':
	case 'v':
	case 'x':
	case 'y':
	case 'z':
	case 'd':
	case 'm':
	  g_fprintf (f, "environment.%c", a->op_type);
	  break;
	case t_xmax:
	  g_fprintf (f, "environment.X");
	  break;
	case t_ymax:
	  g_fprintf (f, "environment.Y");
	  break;
	case t_zmax:
	  g_fprintf (f, "environment.Z");
	  break;
	case t_mmax:
	  g_fprintf (f, "environment.M");
	  break;
	case '!':
	case '~':
	  fputc (a->op_type, f);
	  CT (0);
	  break;
	case t_sign:
	  fputc ('_', f);
	  CT (0);
	  break;
	case t_sin:
	  g_fprintf (f, "p_sin(");
	  CT (0);
	  fputc (')', f);
	  break;
	case t_cos:
	  g_fprintf (f, "p_cos(");
	  CT (0);
	  fputc (')', f);
	  break;
	case t_tan:
	  g_fprintf (f, "p_tan(");
	  CT (0);
	  fputc (')', f);
	  break;
	case t_sqr:
	  g_fprintf (f, "p_sqr(");
	  CT (0);
	  fputc (')', f);
	  break;
	case t_ctl:
	  g_fprintf (f, "p_ctl(");
	  CT (0);
	  fputc (')', f);
	  break;
	case t_get:
	  g_fprintf (f, "p_get(");
	  CT (0);
	  fputc (')', f);
	  break;
	case t_get0:
	  g_fprintf (f, "p_get0(");
	  CT (0);
	  fputc (')', f);
	  break;
	case t_abs:
	  g_fprintf (f, "p_abs(");
	  CT (0);
	  fputc (')', f);
	  break;
	case t_c2d:
	  g_fprintf (f, "p_c2d(");
	  CT (0);
	  fputc (',', f);
	  CT (1);
	  fputc (')', f);
	  break;
	case t_c2m:
	  g_fprintf (f, "p_c2m(");
	  CT (0);
	  fputc (',', f);
	  CT (1);
	  fputc (')', f);
	  break;
	case t_r2x:
	  g_fprintf (f, "p_r2x(");
	  CT (0);
	  fputc (',', f);
	  CT (1);
	  fputc (')', f);
	  break;
	case t_r2y:
	  g_fprintf (f, "p_r2y(");
	  CT (0);
	  fputc (',', f);
	  CT (1);
	  fputc (')', f);
	  break;
	case '/':
	  g_fprintf (f, "p_div(");
	  CT (0);
	  fputc (',', f);
	  CT (1);
	  fputc (')', f);
	  break;
	case '%':
	  g_fprintf (f, "p_mod(");
	  CT (0);
	  fputc (',', f);
	  CT (1);
	  fputc (')', f);
	  break;
	case t_put:
	  g_fprintf (f, "p_put(");
	  CT (0);
	  fputc (',', f);
	  CT (1);
	  fputc (')', f);
	  break;
	case t_put0:
	  g_fprintf (f, "p_put0(");
	  CT (0);
	  fputc (',', f);
	  CT (1);
	  fputc (')', f);
	  break;
	case t_map:
	  g_fprintf (f, "p_map(");
	  CT (0);
	  fputc (',', f);
	  CT (1);
	  fputc (')', f);
	  break;
	case t_add:
	  g_fprintf (f, "p_add(");
	  CT (0);
	  fputc (',', f);
	  CT (1);
	  fputc (',', f);
	  CT (2);
	  fputc (')', f);
	  break;
	case t_mix:
	  g_fprintf (f, "p_mix(");
	  CT (0);
	  fputc (',', f);
	  CT (1);
	  fputc (',', f);
	  CT (2);
	  fputc (',', f);
	  CT (3);
	  fputc (')', f);
	  break;
	case t_sub:
	  g_fprintf (f, "p_sub(");
	  CT (0);
	  fputc (',', f);
	  CT (1);
	  fputc (',', f);
	  CT (2);
	  fputc (')', f);
	  break;
	case t_min:
	  g_fprintf (f, "p_min(");
	  CT (0);
	  fputc (',', f);
	  CT (1);
	  fputc (')', f);
	  break;
	case t_max:
	  g_fprintf (f, "p_max(");
	  CT (0);
	  fputc (',', f);
	  CT (1);
	  fputc (')', f);
	  break;
	case t_rnd:
	  g_fprintf (f, "p_rnd(");
	  CT (0);
	  fputc (',', f);
	  CT (1);
	  fputc (')', f);
	  break;
	case t_dif:
	  g_fprintf (f, "p_dif(");
	  CT (0);
	  fputc (',', f);
	  CT (1);
	  fputc (')', f);
	  break;
	case t_val:
	  g_fprintf (f, "p_val(");
	  CT (0);
	  fputc (',', f);
	  CT (1);
	  fputc (',', f);
	  CT (2);
	  fputc (')', f);
	  break;
	case t_src:
	  g_fprintf (f, "p_src(");
	  CT (0);
	  fputc (',', f);
	  CT (1);
	  fputc (',', f);
	  CT (2);
	  fputc (')', f);
	  break;
	case t_srcr:
	  g_fprintf (f, "p_src_r(");
	  CT (0);
	  fputc (',', f);
	  CT (1);
	  fputc (')', f);
	  break;
	case t_srcg:
	  g_fprintf (f, "p_src_g(");
	  CT (0);
	  fputc (',', f);
	  CT (1);
	  fputc (')', f);
	  break;
	case t_srcb:
	  g_fprintf (f, "p_src_b(");
	  CT (0);
	  fputc (',', f);
	  CT (1);
	  fputc (')', f);
	  break;
	case t_srca:
	  g_fprintf (f, "p_src_a(");
	  CT (0);
	  fputc (',', f);
	  CT (1);
	  fputc (')', f);
	  break;
	case t_rad:
	  g_fprintf (f, "p_rad(");
	  CT (0);
	  CT (1);
	  fputc (',', f);
	  CT (2);
	  fputc (')', f);
	  break;
	case '?':
	  CT (0);
	  fputc ('?', f);
	  CT (1);
	  fputc (':', f);
	  CT (2);
	  break;
	case t_scl:
	  g_fprintf (f, "p_scl(");
	  CT (0);
	  fputc (',', f);
	  CT (1);
	  fputc (',', f);
	  CT (2);
	  fputc (',', f);
	  CT (3);
	  fputc (',', f);
	  CT (4);
	  fputc (')', f);
	  break;
	case t_cnv:
	  g_fprintf (f, "p_cnv(");
	  CT (0);
	  fputc (',', f);
	  CT (1);
	  fputc (',', f);
	  CT (2);
	  fputc (',', f);
	  CT (3);
	  fputc (',', f);
	  CT (4);
	  fputc (',', f);
	  CT (5);
	  fputc (',', f);
	  CT (6);
	  fputc (',', f);
	  CT (7);
	  fputc (',', f);
	  CT (8);
	  fputc (',', f);
	  CT (9);
	  fputc (')', f);
	  break;
	case t_comma:
	  CT (0);
	  fputc (',', f);
	  CT (1);
	  break;
	default:
	  g_fprintf (stderr, "%s %i !\n",
		     _("Unhandled operator"), a->op_type);
	  break;
	}
      fputc (')', f);
    }
}


/*
 *  The native plug-in is created by copying a plug-in skeleton file
 *  and inserting code relevant to the filter at specially marked
 *  locations.
 *
 *  Format of the skeleton file:
 *
 *  The skeleton file should be like a normal C source with special
 *  tokens that is copied from (skeleton) out (output).
 *  Whenever a special token is found in the copy process, it is
 *  substituted by the action of a corresponding function that emits
 *  code to (output).  When this function is done, the copy process
 *  continues until end-of-file is reached.
 *
 *  Tokens:
 *    $$C -> emit plug-in constants
 *    $$F -> emit filter data dump
 *    $$R -> emit channel code for R channel
 *    $$G -> emit channel code for G channel
 *    $$B -> emit channel code for B channel
 *    $$A -> emit channel code for A channel
 */

typedef enum ACTION_CODE {
  ACT_ERROR = -1,
  ACT_EMIT_C, /* constants */
  ACT_EMIT_F, /* filter data */
  ACT_EMIT_R, /* channel 'R' code */
  ACT_EMIT_G, /* channel 'G' code */
  ACT_EMIT_B, /* channel 'B' code */
  ACT_EMIT_A, /* channel 'A' code */
  ACT_DONE
} ActionCode;

static ActionCode
copy_skeleton_part (FILE * i,
		    FILE * o)
{
  guint dollar = 0;

  while (!feof (i))
    {
      gint c;

      if ((c = fgetc(i)) == '$')
	{
	  if (++dollar == 3)
	    {
	      fputc ('$', o);
	      dollar--;
	    }
	}
      else if (c == EOF)
	{
	  return ACT_ERROR;
	}
      else if (dollar == 2)
	{
	  switch (c)
	    {
	    case 'C':
	      return ACT_EMIT_C;
	    case 'F':
	      return ACT_EMIT_F;
	    case 'R':
	      return ACT_EMIT_R;
	    case 'G':
	      return ACT_EMIT_G;
	    case 'B':
	      return ACT_EMIT_B;
	    case 'A':
	      return ACT_EMIT_A;
	    default:
	      return ACT_ERROR;
	    }
	}
      else
	{
	  while (dollar)
	    {
	      fputc ('$', o);
	      dollar--;
	    }
	  if (c == '\n')
	    {
	      ++line;
	      ++sline;
	    }
	  fputc (c, o);
	}
    }
  return ACT_DONE;
}


static gboolean
dump_filter (FILE *             f,
	     const UserFilter * filter,
	     const gboolean     strip)
{
  GError * err = NULL;
  gchar *  data;
  gchar *  p;
  gsize    len;

  data = g_key_file_to_data ((GKeyFile *)filter, &len, &err);
  if (err != NULL)
    {
      g_fprintf (stderr, "%s: %s\n",
		 _("Failed to write filter data"), err->message);
      g_error_free (err);
      g_free (data);
      return FALSE;
    }
  if (strip)
    {
      UserFilter * tmp;
      gboolean     result;

      if (verbose)
	{
	  g_fprintf (stderr,
		     "%s (%u %s) ...\n",
		     _("Stripping filter"), 1 + strlen (data), _("bytes"));
	}
      tmp = guf_filter_new ();
      g_key_file_load_from_data ((GKeyFile *)tmp,
				 data, strlen (data)+1,
				 G_KEY_FILE_KEEP_TRANSLATIONS, &err);
      g_clear_error (&err);
      g_key_file_remove_group ((GKeyFile *)tmp, "Code", &err);
      g_clear_error (&err);

      g_free (data);
      result = dump_filter (f, tmp, FALSE);
      guf_filter_free (tmp);
      return result;
    }
  if (verbose)
    {
      g_fprintf (stderr,
		 "%s (%u %s) ...\n",
		 _("Writing filter"), 1 + strlen (data), _("bytes"));
    }
  g_fprintf (f,
	     "\n"
	     "#line %u \"%s\"\n"
	     "  \"", line++, outfile_name);
  for (p=data; *p; ++p)
    {
      switch (*p)
	{
	case '\n':
	  g_fprintf (f,
		     "\\n\"\n"
		     "  \"");
	  ++line;
	  break;
	case '\"':
	case '\'':
	  fputc ('\\', f);
	  fputc (*p, f);
	  break;
	default:
	  fputc (*p, f);
	  break;
	}
    }
  ++line;
  g_free (data);
  g_fprintf (f,
	     "\";\n"
	     "#line %u \"%s\"\n", sline++, skeleton_name);
  ++line;
  return TRUE;
}


static gboolean
write_source (const UserFilter * filter,
	      FILE *             s,
	      FILE *             f)
{
  ActionCode a;

  g_fprintf (f, "#line %u \"%s\"\n", sline, skeleton_name);
  do
    {
      a = copy_skeleton_part(s, f);

      switch (a)
	{
	case ACT_EMIT_C:
	  {
	    gchar * procedure_name;
	    gchar * data_key_vals;
	    gchar * plugin_name;
	    gchar * title;

	    title = guf_filter_get_title (filter);
	    procedure_name = make_pdb_name ("gimp_userfilter", title, '_');
	    data_key_vals = make_pdb_name ("plug_in_userfilter", title, '_');
	    plugin_name = make_pdb_name ("gimp-userfilter", title, '-');

	    g_fprintf (f, "#line %u \"%s\"\n", line, outfile_name);
	    g_fprintf (f, "#define PROCEDURE_NAME               \"%s\"\n", procedure_name);
	    g_fprintf (f, "#define DATA_KEY_VALS                \"%s\"\n", data_key_vals);
	    g_fprintf (f, "#define PLUGIN_NAME                  \"%s\"\n", plugin_name);
	    g_fprintf (f, "#define GETTEXT_PACKAGE              \"%s\"\n", GETTEXT_PACKAGE);
	    g_fprintf (f, "#define LOCALEDIR                    \"%s\"\n", LOCALEDIR);
	    g_fprintf (f, "#define HAVE_BIND_TEXTDOMAIN_CODESET %u\n", HAVE_BIND_TEXTDOMAIN_CODESET);
	    g_fprintf (f, "#line %u \"%s\"\n", sline++, skeleton_name);
	    line += 5;

	    g_free (plugin_name);
	    g_free (data_key_vals);
	    g_free (procedure_name);
	    g_free (title);
	  }
	  break;
	case ACT_EMIT_F:
	  if (!dump_filter (f, filter, strip))
	    {
	      return FALSE;
	    }
	  break;
	case ACT_EMIT_R:
	case ACT_EMIT_G:
	case ACT_EMIT_B:
	case ACT_EMIT_A:
	  {
	    CodeTree * function;
	    gchar *    source;

	    g_fprintf (f,
		       "#line %u \"%s\"\n"
		       "  return p_min (p_max (0, ", line, outfile_name);
	    if ((source = guf_filter_get_src (filter, (guint)a - (guint)ACT_EMIT_R)) == NULL)
	      {
		return FALSE;
	      }
	    if ((function = get_code_tree (source)) == NULL)
	      {
		g_free (source);
		return FALSE;
	      }
	    if (optimize)
	      {
		static const gchar RGBA[N_CHANNELS] = {
		  'R', 'G', 'B', 'A'
		};

		if (verbose)
		  {
		    g_fprintf (stderr,
			       "%s '%c': %u %s -> ",
			       _("Optimizing channel"), RGBA[(guint)a - (guint)ACT_EMIT_R],
			       count_nodes (function), _("nodes"));
		  }
		optimize_code_tree (function, (guint)a - (guint)ACT_EMIT_R, TRUE, FALSE, NULL);
		if (verbose)
		  {
		    g_fprintf (stderr,
			       "%u %s\n",
			       count_nodes (function), _("nodes"));
		  }
	      }
	    calc_tree (f, function);
	    free_code_tree (function);
	    g_free (source);
	    g_fprintf (f,
		       "), 255);\n"
		       "#line %u \"%s\"", sline, skeleton_name);
	    line += 3;
	  }
	  break;
	default:
	  break;
	}
    }
  while ( (a != ACT_DONE) && (a != ACT_ERROR) );

  return (a == ACT_DONE);
}


static gboolean
compile (const gchar * const infile,
	 const gchar * const skeleton,
	 const gchar * const outfile)
{
  UserFilter * filter;
  gboolean     result = FALSE;
  FILE *       s;
  FILE *       f = NULL;

  if ((s = g_fopen (skeleton, "r")) == NULL)
    {
      g_fprintf (stderr, "%s: %s\n",
		 _("Unable to find skeleton file"), skeleton);
      return result;
    }
  filter = guf_filter_new ();
  if (guf_read (infile, GUF_PROTOCOL, filter))
    {
      if (outfile)
	{
	  if (!(f = g_fopen (outfile, "w")))
	    {
	      goto fail;
	    }
	}
      result = write_source (filter, s, !outfile ? stdout : f);
      if (outfile)
	{
	  fclose (f);
	}
    }
 fail:
  fclose (s);
  guf_filter_free (filter);
  return result;
}


static void
print_help (void)
{
  g_fprintf (stderr,
	     _("Usage: %s [option] <filter%s>\n"
	       "\n"
	       "Create C source code for a native Gimp plug-in from a User Filter file (%s).\n"
	       "\n"
	       "Options:\n"
	       "  -h | --help :\n"
	       "    Display this help.\n"
	       "  -o name | --output=name :\n"
	       "    Write C source to specified output file instead of stdout.\n"
	       "  -s name | --skeleton=name :\n"
	       "    Use specified skeleton file as a plug-in template instead\n"
	       "    of the default.\n"
	       "  -O | --optimize :\n"
	       "    Optimize the intermediate code for the channel functions.\n"
	       "  -S | --strip :\n"
	       "    Strip comments and code from embedded filter dump.\n"
	       "  -v | --verbose :\n"
	       "    Increase verbosity level.\n"
	       "  -V | --version :\n"
	       "    Display version information.\n" "\n"),
	     g_get_prgname (), EXT_GUF, EXT_GUF);
}


static inline void
set_prg_name (const gchar * const path)
{
  gchar * t;

  t = g_path_get_basename (path);
  g_set_prgname (t);
  g_free (t);
}


gint
main (gint     argc,
      gchar ** argv)
{
  bindtextdomain (GETTEXT_PACKAGE, LOCALEDIR);
#ifdef HAVE_BIND_TEXTDOMAIN_CODESET
  bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
#endif
  textdomain (GETTEXT_PACKAGE);

  set_prg_name (argv[0]);

  while (TRUE)
    {
      static struct option long_options[] = {
	/* name      arg,   val,  short */
	{"help",     FALSE, NULL, 'h'},
	{"output",   TRUE,  NULL, 'o'},
	{"optimize", FALSE, NULL, 'O'},
	{"skeleton", TRUE,  NULL, 's'},
	{"strip",    FALSE, NULL, 'S'},
	{"verbose",  FALSE, NULL, 'v'},
	{"version",  FALSE, NULL, 'V'},
	{NULL,       FALSE, NULL, '\0'}
      };
      gint option_index = 0;
      gint c;

      if ((c = getopt_long (argc, argv, "ho:Os:SvV",
			    long_options, &option_index)) < 0)
	{
	  break;
	}

      switch (c)
	{
	case '?':
	case 'h':
	  print_help();
	  return ((c == 'h') ? EXIT_SUCCESS : EXIT_FAILURE);
	case 'o':
	  outfile_name = optarg;
	  break;
	case 'O':
	  optimize = TRUE;
	  break;
	case 's':
	  skeleton_name = optarg;
	  break;
	case 'S':
	  strip = TRUE;
	  break;
	case 'v':
	  ++verbose;
	  break;
	case 'V':
	  g_fprintf (stderr, "uf2c v" PACKAGE_VERSION "\n");
	  return EXIT_SUCCESS;
	}

    }

  if (optind < argc)
    {
      if (!compile (argv[optind], skeleton_name, outfile_name))
	{
	  return EXIT_FAILURE;
	}
    }
  else
    {
      g_fprintf (stderr, "No input file.\n");
      return EXIT_FAILURE;
    }

  return EXIT_SUCCESS;
}
