/*
 * (C) 1997,2003  Jens Ch. Restemeier <jrestemeier@currantbun.com>
 *
 * (C) 2008       Torsten Neuer <tneuer@inwise.de>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *                          -----------
 *
 *  This file contains the generic dialog and related data for all
 *  user filter that run in stand-alone mode (either through the user
 *  filter plugin itself or as compiled native plugin).
 *
 *
 *  $Id: userfilter.c,v 1.21 2008/06/19 16:03:43 tneuer Exp $
 *
 *  $Log: userfilter.c,v $
 *  Revision 1.21  2008/06/19 16:03:43  tneuer
 *  Added popup-menus to notebook widget
 *
 *  Revision 1.20  2008/06/19 13:04:25  tneuer
 *  Group gtk_wdget_show () with gtk_widget_show_all () where applicable
 *
 *  Revision 1.19  2008/06/18 09:08:15  tneuer
 *  Add reset button for edge mode in control panel.
 *
 *  Revision 1.18  2008/06/17 14:43:41  tneuer
 *  Added editor component for default edge mode
 *
 *  Revision 1.17  2008/06/17 13:53:14  tneuer
 *  Bugfix: Correctly set/toggle map tooltips
 *
 *  Revision 1.16  2008/06/17 07:59:19  tneuer
 *  Added tooltip for edge-mode control on filter control panel
 *
 *  Revision 1.15  2008/06/16 16:47:56  tneuer
 *  Bugfix: do not overwrite edge mode on filter re-call
 *
 *  Revision 1.14  2008/06/16 16:34:45  tneuer
 *  Added guf_filter_(get|set)_edge_mode () access functions
 *
 *  Revision 1.13  2008/06/16 10:07:24  tneuer
 *  Changed default edge mode to "smear" to match default behaviour of FF
 *
 *  Revision 1.12  2008/06/13 12:45:50  tneuer
 *  Add support for different edge modes for the pixel fetcher
 *
 *  Revision 1.11  2008/06/11 17:52:50  tneuer
 *  Removed assertions not required anymore.
 *  Pointers to const data are const now, too.
 *
 *  Revision 1.10  2008/06/09 12:05:38  tneuer
 *  Reduced box widget nesting in dialog (dropped outer vbox)
 *  Last filter filename now dynamically stored
 *
 *  Revision 1.9  2008/06/06 13:03:53  tneuer
 *  No more help button when running filters in stand-alone mode
 *
 *  Revision 1.8  2008/06/06 12:27:34  tneuer
 *  Spinbuttons for controls now snap to nearest increment if necessary
 *
 *  Revision 1.7  2008/06/06 11:20:34  tneuer
 *  Handle "Control %u::Step" stettings in filter control panel
 *
 *  Revision 1.6  2008/05/28 10:01:20  tneuer
 *  Using GtkPaned to divide preview and controls
 *
 *  Revision 1.5  2008/05/27 15:09:23  tneuer
 *  Bugfix: development interface overwrites previous control values on re-run
 *
 *  Revision 1.4  2008/05/26 13:03:17  tneuer
 *  Only display random seed reset widget if filter uses rnd() function
 *
 *  Revision 1.3  2008/05/21 12:30:49  tneuer
 *  Minor code cleanup.
 *
 *  Revision 1.2  2008/05/19 10:55:23  tneuer
 *  Code cleanup: merge duplicate code in plug-in dialog(s)
 *
 *  Revision 1.1.1.1  2008/04/27 13:19:05  tneuer
 *  initial import
 *
 *  Revision 1.1  2008/04/16 11:28:04  tneuer
 *  Initial revision
 *
 */

#include "plugin-conf.h"

#include <string.h>
#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

#define __USERFILTER_INTERNAL__ TRUE

#include "userfilter.h"
#include "checksum.h"
#include "render.h"
#include "guf.h"


FilterEnvironment environment;	 /* pixel environment of the renderer */
guint32           drawable_hash; /* default random seed for the image */
UserFilter *      filter = NULL; /* filter data */
FilterValues      control = {	 /* last filter control values */
  { 0, 0, 0, 0, 0, 0, 0, 0 },
  GIMP_PIXEL_FETCHER_EDGE_SMEAR
};

GtkWidget * dialog_dialog  = NULL;
GtkWidget * dialog_preview = NULL;
GtkWidget * dialog_edge    = NULL;
GtkWidget * dialog_rnd     = NULL;
GtkWidget * dialog_sreset[N_CONTROLS] = {
  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
};
GtkObject * dialog_slider[N_CONTROLS] = {
  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
};


gboolean
_filter_uses_rnd (const UserFilter * const filter)
{
  guint i;

  if (filter != NULL)
    {
      for (i=0; i<N_CHANNELS; i++)
	{
	  gchar * src;
	  
	  src = guf_filter_get_src (filter, i);
	  if (strstr (src, "rnd"))
	    {
	      g_free (src);
	      return TRUE;
	    }
	  g_free (src);
	}
    }
  return FALSE;
}


static void
cb_control_slider_update (const GtkWidget * widget,
			  const gpointer    data)
{
  const guint ctl = GPOINTER_TO_INT (data);

  gtk_spin_button_update (GIMP_SCALE_ENTRY_SPINBUTTON (widget));
  control.value[ctl] =
    gtk_adjustment_get_value (GIMP_SCALE_ENTRY_SCALE_ADJ (widget));
}


static void
cb_control_slider_reset (const GtkWidget * widget,
			 const gpointer    data)
{
  const gint      ctl = GPOINTER_TO_INT (data);
  GtkAdjustment * adj;

  adj = GIMP_SCALE_ENTRY_SCALE_ADJ (dialog_slider[ctl]);
  gtk_adjustment_set_value (adj, guf_filter_get_ctl_preset (filter, ctl));
  gtk_adjustment_value_changed (adj);
}


static void
cb_control_edge_mode_reset (const GtkWidget * widget,
			    const gpointer    data)
{
  gtk_combo_box_set_active (GTK_COMBO_BOX (data),
			    control.edge_mode = guf_filter_get_edge_mode (filter));
}


static void
cb_control_edge_mode_update (const GtkWidget * widget,
			     const gpointer    data)
{
  control.edge_mode = gtk_combo_box_get_active (GTK_COMBO_BOX (widget));
}


static void
cb_control_random_reset (const GtkWidget * widget,
			 const gpointer    data)
{
  GtkAdjustment * adj;

  adj = GIMP_RANDOM_SEED_SPINBUTTON_ADJ (data);
  gtk_adjustment_set_value (adj,
			    compute_drawable_checksum (environment.drawable));
  gtk_adjustment_value_changed (adj);
}


static void
cb_preview_invalidated (GimpDrawable * drawable,
			GimpPreview *  preview)
{
  render (drawable, preview, FALSE, NULL);
}


void
_uf_slider_hide (const guint slider)
{
  GtkObject * scale_entry = dialog_slider[slider];

  gtk_widget_hide (GIMP_SCALE_ENTRY_LABEL (scale_entry));
  gtk_widget_hide (GIMP_SCALE_ENTRY_SCALE (scale_entry));
  gtk_widget_hide (GIMP_SCALE_ENTRY_SPINBUTTON (scale_entry));
  gtk_widget_hide (dialog_sreset[slider]);
}


GtkWidget *
_uf_build_preview (GtkPaned *     box,
		   GimpDrawable * drawable)
{
  GtkWidget * preview;

  preview = gimp_zoom_preview_new (drawable);
  gtk_paned_pack1 (box, preview, FALSE, FALSE);
  gtk_widget_show (preview);
  g_signal_connect_swapped (preview, "invalidated",
			    G_CALLBACK (cb_preview_invalidated), drawable);
  return preview;
}


GtkWidget *
_uf_notebook_tab (const GtkWidget *   notebook,
		  const gchar * const icon,
		  const gchar * const text,
		  const gchar * const tooltip)
{
  GtkWidget * vbox;
  GtkWidget * hbox;
  GtkWidget * label;

  vbox = gtk_vbox_new (FALSE, 5);
  {
    hbox = gtk_hbox_new (FALSE, 5);
    {
      label = gtk_image_new_from_stock (icon, GTK_ICON_SIZE_MENU);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);

      label = gtk_label_new (text);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
    }
    gtk_widget_set_tooltip_text (hbox, tooltip);
    gtk_widget_show_all (hbox);
  }
  gtk_notebook_append_page_menu (GTK_NOTEBOOK (notebook), vbox, hbox, gtk_label_new(text));
  gtk_widget_show (vbox);

  return vbox;
}


static inline void
uf_create_control_pair (GtkTable *  table,
			const guint map)
{
  gboolean map_enable   = guf_filter_get_map_enabled (filter, map);
  gchar *  ctl_label[2] = { NULL, NULL };
  gchar *  tooltip[2]   = { NULL, NULL };
  gchar *  map_label    = NULL;
  gchar *  map_tooltip  = NULL;
  guint    ctl;
  guint    i;

  if (map_enable)
    {
      map_label   = guf_filter_get_locale_map_label (filter, map);
      map_tooltip = guf_filter_get_locale_map_tooltip (filter, map);
    }
  else
    {
      ctl_label[0] = guf_filter_get_locale_ctl_label (filter, 2 * map);
      ctl_label[1] = guf_filter_get_locale_ctl_label (filter, 2 * map + 1);
      tooltip[0]   = guf_filter_get_locale_ctl_tooltip (filter, 2 * map);
      tooltip[1]   = guf_filter_get_locale_ctl_tooltip (filter, 2 * map + 1);
    }

  for (i = 0, ctl = 2 * map; i < 2; ctl++, i++)
    {
      const guint step  = guf_filter_get_ctl_step (filter, ctl);
      const guint upper = 255 - (255 % step);

      dialog_slider[ctl] = gimp_scale_entry_new (table, 0, ctl,	         /* attached to table,col,row */
						 map_enable ?
						 (i == 0 ? map_label : "") :
						 ctl_label[i],
						 300,	                 /* scale width */
						 60,	                 /* spinbutton width */
						 control.value[ctl],     /* scale/spinbutton value */
						 0, upper,               /* scale/spinbutton bounds */
						 step, 0,                /* scale/spinbutton increment */
						 0,	                 /* decimal digits */
						 TRUE, 0, upper,
						 map_enable ? map_tooltip : tooltip[i],
						 "");
      g_signal_connect (GTK_OBJECT (dialog_slider[ctl]), "value_changed",
			G_CALLBACK (cb_control_slider_update),
			GINT_TO_POINTER (ctl));
      g_signal_connect_swapped (GTK_OBJECT (dialog_slider[ctl]),
				"value_changed",
				G_CALLBACK (gimp_preview_invalidate),
				dialog_preview);
      gtk_spin_button_set_snap_to_ticks (GIMP_SCALE_ENTRY_SPINBUTTON (dialog_slider[ctl]), TRUE);

      dialog_sreset[ctl] = gtk_button_new_from_stock (GIMP_STOCK_RESET);
      gtk_table_attach (table, dialog_sreset[ctl],
			3, 4, ctl, ctl + 1,
			GTK_SHRINK | GTK_FILL, GTK_SHRINK | GTK_FILL, 0, 0);
      gtk_widget_set_tooltip_text (dialog_sreset[ctl],
				   _("Reset this control to its default value..."));
      g_signal_connect (GTK_OBJECT (dialog_sreset[ctl]), "clicked",
			G_CALLBACK (cb_control_slider_reset),
			GINT_TO_POINTER (ctl));

      if (!map_enable &&
	  !guf_filter_get_ctl_enabled(filter, ctl))
	{
	  _uf_slider_hide (ctl);
	}
      else
	{
	  gtk_widget_show (dialog_sreset[ctl]);
	}
    }
  g_free (map_label);
  g_free (map_tooltip);
  g_free (ctl_label[0]);
  g_free (ctl_label[1]);
  g_free (tooltip[0]);
  g_free (tooltip[1]);
}


static inline GtkWidget *
build_control_edge_mode (const GtkBox * vbox)
{
  GtkWidget * hbox;
  GtkWidget * combo;
  GtkWidget * button;
  GtkWidget * label;

  hbox = gtk_hbox_new (FALSE, 5);
  gtk_container_set_border_width (GTK_CONTAINER (hbox), 5);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

  label = gtk_label_new (_("Edge mode:"));
  gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 5);

  combo = gtk_combo_box_new_text ();
  gtk_box_pack_start (GTK_BOX (hbox), combo, TRUE, TRUE, 0);
  gtk_combo_box_append_text (GTK_COMBO_BOX (combo), _("None"));
  gtk_combo_box_append_text (GTK_COMBO_BOX (combo), _("Wrap"));
  gtk_combo_box_append_text (GTK_COMBO_BOX (combo), _("Smear"));
  gtk_combo_box_append_text (GTK_COMBO_BOX (combo), _("Black"));
  gtk_combo_box_append_text (GTK_COMBO_BOX (combo), _("Background"));
  gtk_combo_box_set_active (GTK_COMBO_BOX (combo), control.edge_mode);
  g_signal_connect (GTK_OBJECT (combo), "changed",
		    G_CALLBACK (cb_control_edge_mode_update), NULL);
  g_signal_connect_swapped (GTK_OBJECT (combo), "changed",
			    G_CALLBACK (gimp_preview_invalidate),
			    dialog_preview);

  button = gtk_button_new_from_stock (GIMP_STOCK_RESET);
  gtk_box_pack_start (GTK_BOX (hbox), button, FALSE, FALSE, 0);
  gtk_widget_set_tooltip_text (button, _("Reset edge mode of the source function to the filter's default..."));
  g_signal_connect (GTK_OBJECT (button), "clicked",
		    G_CALLBACK (cb_control_edge_mode_reset), combo);

  gtk_widget_set_tooltip_text (hbox, _("Change the behaviour of the image source functions for off-image pixels..."));
  gtk_widget_show_all (hbox);

  return combo;
}


static inline GtkWidget *
build_control_rnd_seeder (GtkBox * vbox)
{
  static gboolean no = FALSE;
  GtkWidget *     hbox;
  GtkWidget *     label;
  GtkWidget *     seeder;
  GtkWidget *     button;

  hbox = gtk_hbox_new (FALSE, 5);
  gtk_container_set_border_width (GTK_CONTAINER (hbox), 5);
  gtk_box_pack_end (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

  label = gtk_label_new (_("Random seed:"));
  gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 5);
  gtk_widget_show (label);

  seeder = gimp_random_seed_new (&environment.drawable_hash, &no);
  gtk_box_pack_start (GTK_BOX (hbox), seeder, TRUE, TRUE, 0);
  g_signal_connect_swapped (GTK_OBJECT (GIMP_RANDOM_SEED_SPINBUTTON (seeder)),
			    "value-changed",
			    G_CALLBACK (gimp_preview_invalidate),
			    dialog_preview);
  gtk_widget_show (seeder);

  button = gtk_button_new_from_stock (GIMP_STOCK_RESET);
  gtk_box_pack_start (GTK_BOX (hbox), button, FALSE, FALSE, 0);
  gtk_widget_set_tooltip_text (button,
			       _("Reset the rnd() seed to the image's default..."));
  g_signal_connect (GTK_OBJECT (button), "clicked",
		    G_CALLBACK (cb_control_random_reset), seeder);
  gtk_widget_show (button);

  if (_filter_uses_rnd (filter))
    {
      gtk_widget_show (hbox);
    }

  return hbox;
}


void
_uf_build_page_filter_control (const GtkWidget * notebook)
{
  GtkWidget * vbox;
  GtkWidget * table;
  guint       map;

  vbox = _uf_notebook_tab (notebook,
			   GIMP_STOCK_TOOL_OPTIONS,
			   _("Filter Control"),
			   _("Change the filter's effect on the image..."));

  /*
   *  display sliders for map () / ctl () functions
   */
  table = gtk_table_new (8, 4, FALSE);
  gtk_table_set_col_spacings (GTK_TABLE (table), 5);
  gtk_table_set_row_spacings (GTK_TABLE (table), 5);
  gtk_container_set_border_width (GTK_CONTAINER (table), 5);
  gtk_box_pack_start (GTK_BOX (vbox), table, TRUE, TRUE, 0);
  for (map = 0; map < N_MAPS; map++)
    {
      uf_create_control_pair (GTK_TABLE (table), map);
    }
  gtk_widget_show (table);

  /*
   *  display selector for the edge-mode of the pixel fetcher
   */
  dialog_edge = build_control_edge_mode (GTK_BOX (vbox));

  /*
   *  display control for image's random seed value
   */
  dialog_rnd = build_control_rnd_seeder (GTK_BOX (vbox));
}


static inline void
uf_build_page_plugin_info (const GtkWidget * notebook)
{
  GtkWidget * vbox;
  GtkWidget * stuf;
  gchar *     buf;
  gchar *     ftitle;
  gchar *     fauthor;
  gchar *     fcopyright;
  gchar *     fdescription;

  /*
   *  Collect display data from internal representation...
   */
  ftitle = guf_filter_get_locale_title (filter);
  fauthor = guf_filter_get_author (filter);
  fcopyright = guf_filter_get_copyright (filter);
  fdescription = guf_filter_get_locale_description (filter);

  buf = g_strdup_printf ("%s %s", _("About"), ftitle);
  vbox = _uf_notebook_tab (notebook,
			   GTK_STOCK_ABOUT,
			   buf, _("See who created this plugin..."));
  g_free (buf);

  stuf = gtk_image_new_from_stock (GIMP_STOCK_WILBER, GTK_ICON_SIZE_DIALOG);
  gtk_box_pack_start (GTK_BOX (vbox), stuf, FALSE, FALSE, 20);


  buf = g_strdup_printf ("<big><b>%s v%d.%d.%d</b></big>",
			 ftitle,
			 guf_filter_get_version_major (filter),
			 guf_filter_get_version_minor (filter),
			 guf_filter_get_version_micro (filter));
  stuf = gtk_label_new (NULL);
  gtk_label_set_justify (GTK_LABEL (stuf), GTK_JUSTIFY_CENTER);
  gtk_label_set_markup (GTK_LABEL (stuf), buf);
  gtk_box_pack_start (GTK_BOX (vbox), stuf, FALSE, FALSE, 20);
  g_free (buf);

  stuf = gtk_label_new (fauthor);
  gtk_label_set_justify (GTK_LABEL (stuf), GTK_JUSTIFY_CENTER);
  gtk_box_pack_start (GTK_BOX (vbox), stuf, FALSE, FALSE, 0);
  stuf = gtk_label_new (fcopyright);
  gtk_box_pack_start (GTK_BOX (vbox), stuf, FALSE, FALSE, 0);

  stuf = gtk_label_new (fdescription);
  gtk_label_set_justify (GTK_LABEL (stuf), GTK_JUSTIFY_CENTER);
  gtk_box_pack_start (GTK_BOX (vbox), stuf, TRUE, TRUE, 0);

#if (PLUGIN_MINOR_VERSION % 2)

  stuf = gtk_hseparator_new ();
  gtk_box_pack_start (GTK_BOX (vbox), stuf, FALSE, FALSE, 10);

  /* odd minor versions are considered as development versions */

  stuf = gtk_label_new (_("This is an unstable development release."));
  gtk_label_set_justify (GTK_LABEL (stuf), GTK_JUSTIFY_CENTER);
  gimp_label_set_attributes (GTK_LABEL (stuf),
			     PANGO_ATTR_STYLE, PANGO_STYLE_ITALIC, -1);
  gtk_box_pack_end (GTK_BOX (vbox), stuf, FALSE, FALSE, 10);

#endif

 gtk_widget_show_all (vbox);
}


static inline void
uf_build_notebook (GtkPaned * box)
{
  GtkWidget * notebook;

  notebook = gtk_notebook_new ();

  gtk_notebook_set_tab_pos (GTK_NOTEBOOK (notebook), GTK_POS_TOP);
  gtk_paned_pack2 (box, notebook, TRUE, FALSE);
  _uf_build_page_filter_control (notebook);
  uf_build_page_plugin_info (notebook);
  gtk_notebook_popup_enable (GTK_NOTEBOOK (notebook));
  gtk_widget_show (notebook);
}


gboolean
userfilter_dialog (GimpDrawable * drawable)
{
  GtkWidget * hbox;
  gboolean    run;
  gchar *     title;

  gimp_ui_init (plugin_name, TRUE);

  title = guf_filter_get_locale_title (filter);
  gimp_dialogs_show_help_button (FALSE);
  dialog_dialog = gimp_dialog_new (title, plugin_name,
				   NULL, 0,
				   gimp_standard_help_func, "",
				   GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
				   GTK_STOCK_OK, GTK_RESPONSE_OK, NULL);
  gtk_dialog_set_alternative_button_order (GTK_DIALOG (dialog_dialog),
					   GTK_RESPONSE_OK,
					   GTK_RESPONSE_CANCEL, -1);
  gimp_window_set_transient (GTK_WINDOW (dialog_dialog));
  g_free (title);

  hbox = gtk_hpaned_new ();
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog_dialog)->vbox), hbox, TRUE, TRUE, 0);
  dialog_preview = _uf_build_preview (GTK_PANED (hbox), drawable);
  uf_build_notebook (GTK_PANED (hbox));
  gtk_widget_show (hbox);

  /*  Show the main containers  */

  gtk_widget_show (dialog_dialog);

  /*  Run the dialog  */

  run = (gimp_dialog_run (GIMP_DIALOG (dialog_dialog)) == GTK_RESPONSE_OK);

  /*  Clean up  */

  gtk_widget_destroy (dialog_dialog);

  return run;
}
